/* diversity-apdb-backend.c -
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <diversity-world.h>
#include <diversity-net.h>
#include <diversity-bard.h>
#include <diversity-ap.h>
#include <string.h>
#include "diversity-apdb.h"

static struct {
	DiversityWorld *world;
	DiversityNet *net;
	gboolean net_activated;

	DiversityApdb *apdb;
	gboolean ready;
	gboolean delay_update;
	gboolean delay_query;

	double last_lon, last_lat;
	double query_lon, query_lat;
} apdb;

static void ap_position(DiversityAp *ap, gdouble lon, gdouble lat)
{
	gdouble len;

	len = DIVERSITY_UNITS(200);
	lon -= len / 2;
	lat -= len / 2;

	diversity_object_geometry_set(DIVERSITY_OBJECT(ap),
			lon, lat, len, len);
}

static void on_query_done(DiversityApdb *db, GSList *aps)
{
	apdb.last_lon = apdb.query_lon;
	apdb.last_lat = apdb.query_lat;

	while (aps) {
		DiversityAp *ap;
		ApDesc *desc = aps->data;

		ap = (DiversityAp *)
			diversity_world_lookup_object(apdb.world,
				DIVERSITY_OBJECT_TYPE_AP, desc->hw_addr);
		if (!ap) {
			ap = diversity_ap_new();
			DIVERSITY_OBJECT(ap)->id = g_strdup(desc->hw_addr);

			diversity_ap_set_ssid(ap, desc->ssid);
			ap_position(ap, desc->lon, desc->lat);

			diversity_world_add_object(apdb.world, DIVERSITY_OBJECT(ap));

			g_debug("new AP %s added", desc->ssid);
		}

		aps = aps->next;
	}
}

static void on_update_done(DiversityApdb *db, gboolean success)
{
	g_debug("update %d", success);
}

static void request_update(void)
{
	DiversityObject *obj;
	ApDesc ap;
	gdouble w, h;

	if (!apdb.ready) {
		apdb.delay_update = TRUE;

		return;
	}

	apdb.delay_update = FALSE;

	obj = (DiversityObject *)
		diversity_net_get_active_ap(apdb.net);
	if (!obj) {
		g_debug("no AP connected");

		return;
	}

	diversity_object_geometry_get(obj, &ap.lon, &ap.lat, &w, &h);
	ap.lon += w / 2;
	ap.lat += h / 2;

	ap.hw_addr = diversity_ap_get_hw_address(DIVERSITY_AP(obj));
	ap.ssid = diversity_ap_get_ssid(DIVERSITY_AP(obj));

	diversity_apdb_request_update(apdb.apdb, &ap, 1);

	g_object_unref(obj);
}

static void request_query(void)
{
	if (!apdb.ready) {
		apdb.delay_query = TRUE;

		return;
	}

	apdb.delay_update = FALSE;

	diversity_apdb_request_query(apdb.apdb,
			apdb.query_lon, apdb.query_lat, 1000.0);
}

static void on_ready(DiversityApdb *db, guint version)
{
	if (!version) {
		apdb.ready = FALSE;

		return;
	}

	g_debug("server ready: update %d, query %d",
			apdb.delay_update, apdb.delay_query);

	apdb.ready = TRUE;

	if (apdb.delay_update)
		request_update();

	if (apdb.delay_query)
		request_query();
}

static void on_geometry_changed(DiversityObject *obj, gdouble lon, gdouble lat, gdouble width, gdouble height)
{
	gdouble len;

	if (!apdb.net_activated) {
		g_debug("no connection");

		return;
	}

	len = DIVERSITY_UNITS(100);

	if ((lon - apdb.last_lon) * (lon - apdb.last_lon) +
	    (lat - apdb.last_lat) * (lat - apdb.last_lat) < len * len)
		return;

	apdb.query_lon = lon;
	apdb.query_lat = lat;

	request_query();
}

static void on_net_state_changed(DiversityNet *net, DiversityNetState state, gpointer user_data)
{
	if (state == DIVERSITY_NET_STATE_ACTIVATED) {
		apdb.net_activated = TRUE;

		diversity_apdb_set_server(apdb.apdb,
				"http://diversity.openmoko.net/php/");

		request_update();
	} else {
		apdb.net_activated = FALSE;

		diversity_apdb_set_server(apdb.apdb, NULL);
	}
}

int apdb_init(DiversityWorld *world, GError **error)
{
	DiversityBard *self;
	DiversityNetState state;

	if (!g_thread_supported())
		g_thread_init(NULL);

	apdb.world = world;
	g_object_ref(world);

	apdb.apdb = diversity_apdb_new(NULL);
	g_signal_connect(apdb.apdb, "update-done", G_CALLBACK(on_update_done), NULL);
	g_signal_connect(apdb.apdb, "query-done", G_CALLBACK(on_query_done), NULL);
	g_signal_connect(apdb.apdb, "ready", G_CALLBACK(on_ready), NULL);

	/* XXX DiversityNet provider must be initialized before APDB? */
	apdb.net = diversity_world_net_get(world);
	g_object_ref(apdb.net);

	g_signal_connect(apdb.net, "state-changed",
			G_CALLBACK(on_net_state_changed), NULL);

	state = diversity_net_get_state(apdb.net);
	on_net_state_changed(apdb.net, state, NULL);

	self = diversity_world_get_self(world);
	g_signal_connect(self, "geometry-changed",
			G_CALLBACK(on_geometry_changed), NULL);

	return TRUE;
}

void apdb_fini(void)
{
	if (apdb.apdb) {
		g_object_unref(apdb.apdb);
		apdb.apdb = NULL;
	}

	if (apdb.net) {
		g_object_unref(G_OBJECT(apdb.net));
		apdb.net = NULL;
	}

	if (apdb.world) {
		g_object_unref(G_OBJECT(apdb.world));
		apdb.world = NULL;
	}
}
