/* diversity-apdb.c - DiversityApdb
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@opeapdboko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <string.h>
#include <libsoup/soup.h>
#include "diversity-apdb.h"

enum {
	READY,
	UPDATE_DONE,
	QUERY_DONE,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_SERVER,
};


#define DIVERSITY_APDB_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_APDB, DiversityApdbPrivate))
typedef struct _DiversityApdbPrivate DiversityApdbPrivate;

struct _DiversityApdbPrivate {
	guint version;
	SoupSession *session;

	gchar *server;
	gchar *http_method;

	SoupMessage *query;
	SoupMessage *update;
};

static guint apdb_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(DiversityApdb, diversity_apdb, G_TYPE_OBJECT);

static void diversity_apdb_dispose(GObject *obj)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(obj);

	if (priv->query) {
		g_object_unref(priv->query);
		priv->query = NULL;
	}

	if (priv->update) {
		g_object_unref(priv->update);
		priv->update = NULL;
	}

	if (priv->session) {
		g_object_unref(priv->session);
		priv->session = NULL;
	}

	((GObjectClass *) diversity_apdb_parent_class)->dispose(obj);
}

static void diversity_apdb_finalize(GObject *obj)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(obj);

	g_free(priv->server);
	g_free(priv->http_method);

	((GObjectClass *) diversity_apdb_parent_class)->finalize(obj);
}

static void diversity_apdb_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	DiversityApdb *apdb = DIVERSITY_APDB(obj);

	switch (prop_id) {
	case PROP_SERVER:
		diversity_apdb_set_server(apdb, g_value_get_string(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_apdb_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec)
{
	DiversityApdb *apdb = DIVERSITY_APDB(obj);
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);

	switch (prop_id) {
	case PROP_SERVER:
		g_value_set_string(value, priv->server);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_apdb_class_init(DiversityApdbClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;

	o_class->dispose = diversity_apdb_dispose;
	o_class->finalize = diversity_apdb_finalize;
	o_class->set_property = diversity_apdb_set_property;
	o_class->get_property = diversity_apdb_get_property;

	apdb_signals[READY] =
		g_signal_new("ready",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityApdbClass, ready),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__UINT,
			     G_TYPE_NONE, 1,
			     G_TYPE_UINT);

	apdb_signals[UPDATE_DONE] =
		g_signal_new("update-done",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityApdbClass, update_done),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__BOOLEAN,
			     G_TYPE_NONE, 1,
			     G_TYPE_BOOLEAN);

	apdb_signals[QUERY_DONE] =
		g_signal_new("query-done",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityApdbClass, query_done),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	g_object_class_install_property(o_class,
					PROP_SERVER,
					g_param_spec_string("server",
						 	    "Server",
							    "The URL of the server",
							    "",
							    G_PARAM_READWRITE));

	g_type_class_add_private(klass, sizeof(DiversityApdbPrivate));
}

static void diversity_apdb_init(DiversityApdb *apdb)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);

	priv->version = 0;
	priv->session = soup_session_async_new();

	priv->http_method = g_strdup("POST");
}

DiversityApdb *diversity_apdb_new(const gchar *url)
{
	DiversityApdb *apdb;

	apdb = g_object_new(DIVERSITY_TYPE_APDB,
			    "server", url,
			    NULL);

	return apdb;
}

/*
 * query: (lat, lon, radius)
 * -> array of (lat, lon, mac, ssid)
 *
 * update: array of (lat, lon, mac, ssid)
 */

gchar *build_method_hello(void)
{
	gchar *body;

	body = soup_xmlrpc_build_method_call("ap.hello", NULL, 0);

	return body;
}

gchar *build_method_query(double lon, double lat, double radius)
{
	gchar *body;
	GValue vals[3] = { { 0 } };

	g_value_init(&vals[0], G_TYPE_DOUBLE);
	g_value_init(&vals[1], G_TYPE_DOUBLE);
	g_value_init(&vals[2], G_TYPE_DOUBLE);

	g_value_set_double(&vals[0], lat);
	g_value_set_double(&vals[1], lon);
	g_value_set_double(&vals[2], radius);

	body = soup_xmlrpc_build_method_call("ap.query", vals, 3);

	//g_debug("Query\n%s\n", body);

	return body;
}

gchar *build_method_update(ApDesc *aps, guint num_aps)
{
	gchar *body;
	GValue val = { 0 };
	GValueArray *array;
	guint i;
	
	array = soup_value_array_new();

	for (i = 0; i < num_aps; i++) {
		GValueArray *ap;

		ap = soup_value_array_new_with_vals(
				G_TYPE_DOUBLE, aps[i].lat,
				G_TYPE_DOUBLE, aps[i].lon,
				G_TYPE_STRING, aps[i].hw_addr,
				G_TYPE_STRING, aps[i].ssid,
				G_TYPE_INVALID, NULL);

		soup_value_array_append(array,
				G_TYPE_VALUE_ARRAY, ap);

		g_value_array_free(ap);
	}

	g_value_init(&val, G_TYPE_VALUE_ARRAY);
	g_value_set_static_boxed(&val, array);

	body = soup_xmlrpc_build_method_call("ap.update", &val, 1);

	g_value_array_free(array);

	//g_debug("Update\n%s\n", body);

	return body;
}

static gboolean parse_response(SoupMessage *msg, GValue *val, GError **error)
{
	if (!SOUP_STATUS_IS_SUCCESSFUL(msg->status_code)) {
		g_set_error(error, 0, 0, "ERROR %d: %s",
				msg->status_code, msg->reason_phrase);

		return FALSE;
	}

	//g_debug("parse response\n%s\n", msg->response_body->data);

	if (!soup_xmlrpc_parse_method_response(
				msg->response_body->data,
				msg->response_body->length,
				val, error)) {
		if (!*error)
			g_set_error(error, 0, 0, "failed to parse response");

		return FALSE;
	}

	return TRUE;
}

static SoupMessage *diversity_apdb_message_new(DiversityApdb *apdb, gchar *body)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);
	SoupMessage *msg;

	if (!body)
		return NULL;

	msg = soup_message_new(priv->http_method, priv->server);
	if (!msg)
		return NULL;

	soup_message_set_request(msg, "text/xml", SOUP_MEMORY_TAKE,
				body, strlen(body));

	return msg;
}

static void on_hello_done(SoupSession *session, SoupMessage *msg, gpointer user_data)
{
	DiversityApdb *apdb = user_data;
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);
	GValue retval = { 0 };
	GError *error = NULL;
	gint ver;

	if (!parse_response(msg, &retval, &error)) {
		g_debug("hello failed: %s", error->message);
		g_error_free(error);

		g_signal_emit(apdb, apdb_signals[READY], 0, 0);

		return;
	}

	ver = g_value_get_int(&retval);
	priv->version = ver;

	g_signal_emit(apdb, apdb_signals[READY], 0,
			ver);

	g_value_unset(&retval);
}

void diversity_apdb_set_server(DiversityApdb *apdb, const gchar *url)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);
	gchar *body;
	SoupMessage *msg;

	soup_session_abort(priv->session);

	g_free(priv->server);
	priv->server = g_strdup(url);
	priv->version = 0;
	g_signal_emit(apdb, apdb_signals[READY], 0, 0);

	if (!url)
		return;

	body = build_method_hello();
	msg = diversity_apdb_message_new(apdb, body);
	if (!msg)
		return;

	soup_session_queue_message(priv->session, msg, on_hello_done, apdb);
}

const gchar *diversity_apdb_get_server(DiversityApdb *apdb)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);

	return priv->server;
}

gboolean diversity_apdb_is_ready(DiversityApdb *apdb)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);

	return (priv->version > 0);
}

static void on_update_done(SoupSession *session, SoupMessage *msg, gpointer user_data)
{
	DiversityApdb *apdb = user_data;
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);
	GValue retval = { 0 };
	GError *error = NULL;

	priv->update = NULL;

	if (!parse_response(msg, &retval, &error)) {
		g_debug("update failed: %s", error->message);
		g_error_free(error);

		g_signal_emit(apdb, apdb_signals[UPDATE_DONE], 0,
				FALSE);

		return;
	}

	g_debug("AP updated");

	g_signal_emit(apdb, apdb_signals[UPDATE_DONE], 0,
			TRUE);
}

gboolean diversity_apdb_request_update(DiversityApdb *apdb, ApDesc *aps, guint num_aps)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);
	gchar *body;

	if (!priv->version)
		return FALSE;

	if (priv->update)
		diversity_apdb_cancel_update(apdb);

	g_return_val_if_fail(aps, FALSE);
	g_return_val_if_fail(num_aps, FALSE);

	body = build_method_update(aps, num_aps);
	priv->update = diversity_apdb_message_new(apdb, body);
	if (!priv->update)
		return FALSE;

	soup_session_queue_message(priv->session, priv->update, on_update_done, apdb);

	return TRUE;
}

void diversity_apdb_cancel_update(DiversityApdb *apdb)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);

	if (priv->update) {
		soup_session_cancel_message(priv->session,
				priv->update, SOUP_STATUS_CANCELLED);

		priv->update = NULL;
	}
}

static gboolean ap_demarshal(GValueArray *array, ApDesc *ap)
{
	if (array->n_values != 4)
		return FALSE;

	if (G_VALUE_TYPE(array->values + 0) != G_TYPE_DOUBLE ||
	    G_VALUE_TYPE(array->values + 1) != G_TYPE_DOUBLE ||
	    G_VALUE_TYPE(array->values + 2) != G_TYPE_STRING ||
	    G_VALUE_TYPE(array->values + 3) != G_TYPE_STRING)
		return FALSE;

	ap->lat = g_value_get_double(array->values);
	ap->lon = g_value_get_double(array->values + 1);
	ap->hw_addr = g_value_get_string(array->values + 2);
	ap->ssid = g_value_get_string(array->values + 3);

	if (!ap->hw_addr)
		return FALSE;

	return TRUE;
}

static void on_query_done(SoupSession *session, SoupMessage *msg, gpointer user_data)
{
	DiversityApdb *apdb = user_data;
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);
	GValueArray *array;
	guint i;
	GValue retval = { 0 };
	GError *error = NULL;
	GSList *aps = NULL;

	priv->query = NULL;

	if (!parse_response(msg, &retval, &error)) {
		g_debug("query failed: %s", error->message);
		g_error_free(error);

		g_signal_emit(apdb, apdb_signals[QUERY_DONE], 0, NULL);

		return;
	}

	array = g_value_get_boxed(&retval);

	g_debug("got %d APs", array->n_values);

	for (i = 0; i < array->n_values; i++) {
		GValueArray *ap;
		ApDesc *desc;
		
		if (G_VALUE_TYPE(array->values + i) != G_TYPE_VALUE_ARRAY) {
			g_debug("AP %d is of type %s", i,
					G_VALUE_TYPE_NAME(array->values + i));

			continue;
		}

		ap = g_value_get_boxed(array->values + i);
		desc = g_new(ApDesc, 1);

		if (!ap_demarshal(ap, desc)) {
			g_debug("failed to demarshal ap");
			g_free(desc);

			continue;
		}

		aps = g_slist_prepend(aps, desc);
	}

	g_signal_emit(apdb, apdb_signals[QUERY_DONE], 0,
			aps);

	g_slist_foreach(aps, (GFunc) g_free, NULL);
	g_slist_free(aps);

	g_value_unset(&retval);
}

gboolean diversity_apdb_request_query(DiversityApdb *apdb, gdouble lon, gdouble lat, gdouble radius)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);
	gchar *body;

	if (!priv->version)
		return FALSE;

	if (priv->query)
		diversity_apdb_cancel_query(apdb);

	body = build_method_query(lon, lat, radius);
	priv->query = diversity_apdb_message_new(apdb, body);
	if (!priv->query)
		return FALSE;

	soup_session_queue_message(priv->session, priv->query, on_query_done, apdb);

	return TRUE;
}

void diversity_apdb_cancel_query(DiversityApdb *apdb)
{
	DiversityApdbPrivate *priv = DIVERSITY_APDB_GET_PRIVATE(apdb);

	if (priv->query) {
		soup_session_cancel_message(priv->session,
				priv->query, SOUP_STATUS_CANCELLED);

		priv->query = NULL;
	}
}
