/* diversity-conn-iface.c - DiversityConnIface
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <stdlib.h>
#include <string.h>
#include <dbus/dbus-glib.h>
#include <time.h>
#include "diversity-connman.h"
#include "diversity-conn-iface.h"

enum {
	NETWORK_FOUND,
	NETWORK_CHANGED,
	AP_ADDED,
	AP_REMOVED,
	AP_CHANGED,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_OBJPATH,
	PROP_STATE,
	PROP_POLICY,
	PROP_SIGNAL,
};


#define DIVERSITY_CONN_IFACE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_CONN_IFACE, DiversityConnIfacePrivate))
typedef struct _DiversityConnIfacePrivate DiversityConnIfacePrivate;

struct _DiversityConnIfacePrivate {
	DBusGConnection *connection;

	gchar *objpath;
	DiversityConnIfaceType type;
	DiversityConnIfaceState state;
	DiversityConnIfacePolicy policy;
	guint signal;

	gchar *driver;
	gchar *vendor;
	gchar *product;

	DiversityConnNetwork *active;
	DiversityConnIPv4 *ipv4;

	GSList *aps;
	guint expiration;

	DBusGProxy *interface;
};

static guint conn_iface_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(DiversityConnIface, diversity_conn_iface, G_TYPE_OBJECT);

static DiversityConnNetwork *diversity_conn_network_new(void)
{
	DiversityConnNetwork *net;

	net = g_slice_new0(DiversityConnNetwork);
	net->last_seen = time(NULL);

	return net;
}

static void network_from_hash(const gchar *key, GValue *val, DiversityConnNetwork *net)
{
	if (strcmp(key, "BSSID") == 0)
		net->bssid = g_value_dup_string(val);
	else if (strcmp(key, "ESSID") == 0)
		net->essid = g_value_dup_string(val);
	else if (strcmp(key, "PSK") == 0)
		net->psk = g_value_dup_string(val);
	else if (strcmp(key, "Security") == 0)
		net->security = g_value_dup_string(val);
	else if (strcmp(key, "Signal") == 0)
		net->signal = g_value_get_uint(val);
	else
		g_debug("unrecognized key %s", key);
}

static DiversityConnNetwork *diversity_conn_network_new_from_hash(GHashTable *hash)
{
	DiversityConnNetwork *net;

	net = diversity_conn_network_new();

	g_hash_table_foreach(hash, (GHFunc) network_from_hash, net);

	return net;
}

static void variant_unset(gpointer data)
{
	g_value_unset(data);
	g_free(data);
}

static GHashTable *diversity_conn_network_hash(DiversityConnNetwork *net)
{
	GHashTable *hash;
	GValue *val;
	struct {
		GType type;
		const gchar *key;
		gchar **val;
	} table[] = {
		{ G_TYPE_STRING,	"BSSID",	&net->bssid	},
		{ G_TYPE_STRING,	"ESSID",	&net->essid	},
		{ G_TYPE_STRING,	"PSK",		&net->psk	},
		{ G_TYPE_INVALID,	NULL,		NULL		}
	}, *p = table;

	hash = g_hash_table_new_full(g_str_hash, g_str_equal,
			NULL, variant_unset);

	while (p->type != G_TYPE_INVALID) {
		if (!*(p->val)) {
			p++;

			continue;
		}

		val = g_new0(GValue, 1);
		g_value_init(val, p->type);
		g_value_set_string(val, *(p->val));

		g_hash_table_insert(hash, (gchar *) p->key, val);

		p++;
	}

	return hash;
}

static void diversity_conn_network_free(DiversityConnNetwork *net)
{
	g_free(net->bssid);
	g_free(net->essid);
	g_free(net->psk);
	g_free(net->security);

	g_slice_free(DiversityConnNetwork, net);
}

static void diversity_conn_iface_dispose(GObject *obj)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(obj);

	if (priv->interface) {
		g_object_unref(priv->interface);
		priv->interface = NULL;
	}

	((GObjectClass *) diversity_conn_iface_parent_class)->dispose(obj);
}

static void diversity_conn_iface_finalize(GObject *obj)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(obj);

	g_slist_foreach(priv->aps, (GFunc) diversity_conn_network_free, NULL);
	g_slist_free(priv->aps);

	if (priv->active)
		diversity_conn_network_free(priv->active);

	g_free(priv->driver);
	g_free(priv->vendor);
	g_free(priv->product);

	g_free(priv->objpath);
	dbus_g_connection_unref(priv->connection);

	g_free(priv->ipv4);

	((GObjectClass *) diversity_conn_iface_parent_class)->finalize(obj);
}

static void diversity_conn_iface_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	DiversityConnIface *cface = DIVERSITY_CONN_IFACE(obj);
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	switch (prop_id) {
	case PROP_OBJPATH:
		g_free(priv->objpath);
		priv->objpath = g_value_dup_string(value);
		break;
	case PROP_POLICY:
		diversity_conn_iface_set_policy(cface,
				g_value_get_int(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_conn_iface_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec)
{
	DiversityConnIface *cface = DIVERSITY_CONN_IFACE(obj);
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	switch (prop_id) {
	case PROP_OBJPATH:
		g_value_set_string(value, priv->objpath);
		break;
	case PROP_STATE:
		g_value_set_int(value, priv->state);
		break;
	case PROP_POLICY:
		g_value_set_int(value, priv->policy);
		break;
	case PROP_SIGNAL:
		g_value_set_uint(value, priv->signal);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_conn_iface_class_init(DiversityConnIfaceClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;

	o_class->dispose = diversity_conn_iface_dispose;
	o_class->finalize = diversity_conn_iface_finalize;
	o_class->set_property = diversity_conn_iface_set_property;
	o_class->get_property = diversity_conn_iface_get_property;

	conn_iface_signals[NETWORK_FOUND] =
		g_signal_new("network-found",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityConnIfaceClass,
					     network_found),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	conn_iface_signals[NETWORK_CHANGED] =
		g_signal_new("network-changed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityConnIfaceClass,
					     network_changed),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	conn_iface_signals[AP_ADDED] =
		g_signal_new("ap-added",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityConnIfaceClass,
					     ap_added),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	conn_iface_signals[AP_REMOVED] =
		g_signal_new("ap-removed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityConnIfaceClass,
					     ap_removed),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	conn_iface_signals[AP_CHANGED] =
		g_signal_new("ap-changed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityConnIfaceClass,
					     ap_changed),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	g_object_class_install_property(o_class,
					PROP_OBJPATH,
					g_param_spec_string(
						"objpath",
						"Object path",
						"Object path",
						"",
						G_PARAM_READWRITE |
						G_PARAM_CONSTRUCT_ONLY));

	g_object_class_install_property(o_class,
					PROP_STATE,
					g_param_spec_int(
						"state",
						"State",
						"State",
						DIVERSITY_CONN_IFACE_STATE_UNKNOWN,
						DIVERSITY_CONN_IFACE_STATE_SHUTDOWN,
						DIVERSITY_CONN_IFACE_STATE_UNKNOWN,
						G_PARAM_READABLE));

	g_object_class_install_property(o_class,
					PROP_POLICY,
					g_param_spec_int(
						"policy",
						"Policy",
						"Policy",
						DIVERSITY_CONN_IFACE_POLICY_UNKNOWN,
						DIVERSITY_CONN_IFACE_POLICY_ASK,
						DIVERSITY_CONN_IFACE_POLICY_UNKNOWN,
						G_PARAM_READWRITE));

	g_object_class_install_property(o_class,
					PROP_SIGNAL,
					g_param_spec_uint(
						"signal",
						"Signal",
						"Signal strength",
						0,
						100,
						0,
						G_PARAM_READABLE));

	g_type_class_add_private(klass, sizeof(DiversityConnIfacePrivate));
}

static void diversity_conn_iface_init(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	priv->objpath = NULL;
	priv->signal = 0;
	priv->type = DIVERSITY_CONN_IFACE_TYPE_UNKNOWN;
	priv->state = DIVERSITY_CONN_IFACE_STATE_UNKNOWN;
	priv->policy = DIVERSITY_CONN_IFACE_POLICY_UNKNOWN;

	priv->expiration = 30;

	priv->ipv4 = g_new0(DiversityConnIPv4, 1);
}

static DiversityConnIfaceType type_string_to_type(const gchar *type)
{
	DiversityConnIfaceType t;

	if (!type)
		return DIVERSITY_CONN_IFACE_TYPE_UNKNOWN;

	if (strcmp(type, "80203") == 0)
		t = DIVERSITY_CONN_IFACE_TYPE_80203;
	else if (strcmp(type, "80211") == 0)
		t = DIVERSITY_CONN_IFACE_TYPE_80211;
	else if (strcmp(type, "wimax") == 0)
		t = DIVERSITY_CONN_IFACE_TYPE_WIMAX;
	else if (strcmp(type, "modem") == 0)
		t = DIVERSITY_CONN_IFACE_TYPE_MODEM;
	else if (strcmp(type, "bluetooth") == 0)
		t = DIVERSITY_CONN_IFACE_TYPE_BLUETOOTH;
	else
		t = DIVERSITY_CONN_IFACE_TYPE_UNKNOWN;

	return t;
}

static DiversityConnIfaceState state_string_to_state(const gchar *state)
{
	DiversityConnIfaceState s;

	if (!state)
		return DIVERSITY_CONN_IFACE_STATE_UNKNOWN;

	switch (state[0]) {
	case 'e': /* "enabled" */
		s = DIVERSITY_CONN_IFACE_STATE_ENABLED;
		break;
	case 'o': /* " off */
		s = DIVERSITY_CONN_IFACE_STATE_OFF;
		break;
	case 'r': /* "ready" */
		s = DIVERSITY_CONN_IFACE_STATE_READY;
		break;
	case 's':
		if (strcmp(state, "scanning") == 0)
			s = DIVERSITY_CONN_IFACE_STATE_SCANNING;
		else if (strcmp(state, "shutdown") == 0)
			s = DIVERSITY_CONN_IFACE_STATE_SHUTDOWN;
		else
			s = DIVERSITY_CONN_IFACE_STATE_UNKNOWN;
		break;
	case 'c':
		if (strcmp(state, "connect") == 0)
			s = DIVERSITY_CONN_IFACE_STATE_CONNECT;
		else if (strcmp(state, "connected") == 0)
			s = DIVERSITY_CONN_IFACE_STATE_CONNECTED;
		else if (strcmp(state, "carrier") == 0)
			s = DIVERSITY_CONN_IFACE_STATE_CARRIER;
		else if (strcmp(state, "configure") == 0)
			s = DIVERSITY_CONN_IFACE_STATE_CONFIGURE;
		else
			s = DIVERSITY_CONN_IFACE_STATE_UNKNOWN;
		break;
	case 'u': /* "unknown" */
	default:
		s = DIVERSITY_CONN_IFACE_STATE_UNKNOWN;
		break;
	}

	return s;
}

static DiversityConnIfacePolicy policy_string_to_policy(const gchar *policy)
{
	DiversityConnIfacePolicy p;

	if (!policy)
		return DIVERSITY_CONN_IFACE_POLICY_UNKNOWN;

	if (strcmp(policy, "off") == 0)
		p = DIVERSITY_CONN_IFACE_POLICY_OFF;
	else if (strcmp(policy, "ignore") == 0)
		p = DIVERSITY_CONN_IFACE_POLICY_IGNORE;
	else if (strcmp(policy, "auto") == 0)
		p = DIVERSITY_CONN_IFACE_POLICY_AUTO;
	else if (strcmp(policy, "ask") == 0)
		p = DIVERSITY_CONN_IFACE_POLICY_ASK;
	else
		p = DIVERSITY_CONN_IFACE_POLICY_UNKNOWN;

	return p;
}

static const gchar *policy_to_policy_string(DiversityConnIfacePolicy policy)
{
	const gchar *p;

	switch (policy) {
	case DIVERSITY_CONN_IFACE_POLICY_OFF:
		p = "off";
		break;
	case DIVERSITY_CONN_IFACE_POLICY_IGNORE:
		p = "ignore";
		break;
	case DIVERSITY_CONN_IFACE_POLICY_AUTO:
		p = "auto";
		break;
	case DIVERSITY_CONN_IFACE_POLICY_ASK:
		p = "ask";
		break;
	case DIVERSITY_CONN_IFACE_POLICY_UNKNOWN:
	default:
		p = "unknown";
		break;
	}

	return p;
}

static void on_signal_changed(DBusGProxy *proxy, guint signal, DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	if (priv->signal != signal) {
		priv->signal = signal;

		g_object_notify(G_OBJECT(cface), "signal");
	}
}

static void on_state_changed(DBusGProxy *proxy, const gchar *state, DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);
	DiversityConnIfaceState s = state_string_to_state(state);

	if (priv->state != s) {
		priv->state = s;

		g_object_notify(G_OBJECT(cface), "state");
	}
}

static void on_policy_changed(DBusGProxy *proxy, const gchar *policy, DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);
	DiversityConnIfaceState p = policy_string_to_policy(policy);

	if (priv->policy != p) {
		priv->policy = p;

		g_object_notify(G_OBJECT(cface), "policy");
	}
}

static void on_network_changed(DBusGProxy *proxy, GHashTable *network, DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	if (priv->active)
		diversity_conn_network_free(priv->active);

	priv->active = diversity_conn_network_new_from_hash(network);

	g_signal_emit(cface, conn_iface_signals[NETWORK_CHANGED], 0,
			priv->active);
}

static gint network_compare(DiversityConnNetwork *a, DiversityConnNetwork *b)
{
	if (!b->bssid)
		return 1;
	else if (!a->bssid)
		return -1;

	return strcmp(a->bssid, b->bssid);
}

static gboolean network_timeout(gpointer data)
{
	DiversityConnIface *cface = data;
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);
	GSList *tmp_list;
	time_t now = time(NULL);

	tmp_list = priv->aps;
	while (tmp_list) {
		DiversityConnNetwork *net = tmp_list->data;
		GSList *next = tmp_list->next;

		if (net->last_seen + priv->expiration <= now) {
			g_signal_emit(cface,
					conn_iface_signals[AP_REMOVED], 0,
					net);
			diversity_conn_network_free(net);
			
			priv->aps = g_slist_delete_link(priv->aps, tmp_list);
		}

		tmp_list = next;
	}

	return (priv->aps != NULL);
}

static void on_network_found(DBusGProxy *proxy, GHashTable *network, DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);
	DiversityConnNetwork *n;
	GSList *old;

	g_signal_emit(cface, conn_iface_signals[NETWORK_FOUND], 0,
			network);

	n = diversity_conn_network_new_from_hash(network);

	old = g_slist_find_custom(priv->aps, n,
			(GCompareFunc) network_compare);
	if (old) {
		diversity_conn_network_free(old->data);
		old->data = n;

		g_signal_emit(cface, conn_iface_signals[AP_CHANGED], 0,
				n);
	} else {
		if (!priv->aps)
			g_timeout_add_seconds(priv->expiration,
					network_timeout, cface);

		priv->aps = g_slist_prepend(priv->aps, n);
		g_signal_emit(cface, conn_iface_signals[AP_ADDED], 0,
				n);
	}
}

static gboolean diversity_conn_iface_bind(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);
	guint signal;
	gchar *state;
	gchar *policy;
	GHashTable *network, *props;
	GType type;

	type = dbus_g_type_get_map("GHashTable",
			G_TYPE_STRING, G_TYPE_VALUE);

	if (dbus_g_proxy_call(priv->interface, "GetProperties", NULL,
				G_TYPE_INVALID,
				type, &props,
				G_TYPE_INVALID)) {
		GValue *val;

		val = g_hash_table_lookup(props, "Type");
		if (val) {
			const gchar *type = g_value_get_string(val);

			priv->type = type_string_to_type(type);
		}

		val = g_hash_table_lookup(props, "Driver");
		if (val)
			priv->driver = g_value_dup_string(val);

		val = g_hash_table_lookup(props, "Vendor");
		if (val)
			priv->vendor = g_value_dup_string(val);

		val = g_hash_table_lookup(props, "Product");
		if (val)
			priv->product = g_value_dup_string(val);

		g_hash_table_destroy(props);
	}

	dbus_g_proxy_add_signal(priv->interface, "SignalChanged",
				G_TYPE_UINT, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(priv->interface, "SignalChanged",
				G_CALLBACK(on_signal_changed), cface, NULL);

	if (dbus_g_proxy_call(priv->interface, "GetSignal", NULL,
				G_TYPE_INVALID,
				G_TYPE_UINT, &signal,
				G_TYPE_INVALID)) {
		priv->signal = signal;
	}

	dbus_g_proxy_add_signal(priv->interface, "StateChanged",
				G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(priv->interface, "StateChanged",
				G_CALLBACK(on_state_changed), cface, NULL);

	if (dbus_g_proxy_call(priv->interface, "GetState", NULL,
				G_TYPE_INVALID,
				G_TYPE_STRING, &state,
				G_TYPE_INVALID)) {
		priv->state = state_string_to_state(state);
		g_free(state);
	}

	dbus_g_proxy_add_signal(priv->interface, "PolicyChanged",
				G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(priv->interface, "PolicyChanged",
				G_CALLBACK(on_policy_changed), cface, NULL);

	if (dbus_g_proxy_call(priv->interface, "GetPolicy", NULL,
				G_TYPE_INVALID,
				G_TYPE_STRING, &policy,
				G_TYPE_INVALID)) {
		priv->policy = policy_string_to_policy(policy);
		g_free(policy);
	}

	dbus_g_proxy_add_signal(priv->interface, "NetworkChanged",
				type, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(priv->interface, "NetworkChanged",
				G_CALLBACK(on_network_changed), cface, NULL);

	if (dbus_g_proxy_call(priv->interface, "GetNetwork", NULL,
				G_TYPE_INVALID,
				type, &network,
				G_TYPE_INVALID)) {
		priv->active =
			diversity_conn_network_new_from_hash(network);

		g_hash_table_destroy(network);
	}

	dbus_g_proxy_add_signal(priv->interface, "NetworkFound",
				type, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(priv->interface, "NetworkFound",
				G_CALLBACK(on_network_found), cface, NULL);

	return TRUE;
}

DiversityConnIface *diversity_conn_iface_new(const gchar *objpath, GError **error)
{
	DiversityConnIface *cface;
	DiversityConnIfacePrivate *priv;
	
	cface = g_object_new(DIVERSITY_TYPE_CONN_IFACE,
			"objpath", objpath,
			NULL);
	priv = DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	priv->connection = dbus_g_bus_get(CONNMAN_DBUS_BUS, error);
	if (!priv->connection) {
		g_object_unref(cface);

		return NULL;
	}

	priv->interface =
		dbus_g_proxy_new_for_name(priv->connection,
					  CONNMAN_DBUS_SERVICE,
					  objpath,
					  CONNMAN_INTERFACE_DBUS_INTERFACE);
	if (!priv->interface) {
		g_set_error(error, 0, 0, "failed to create proxy");

		g_object_unref(cface);

		return NULL;
	}

	if (!diversity_conn_iface_bind(cface)) {
		g_set_error(error, 0, 0,
				"failed to bind interface %s", objpath);
		g_object_unref(cface);

		return NULL;
	}

	return cface;
}

DiversityConnIfaceType diversity_conn_iface_get_iface_type(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	return priv->type;
}

const gchar *diversity_conn_iface_get_object_path(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	return priv->objpath;
}

const gchar *diversity_conn_iface_get_driver(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	return priv->driver;
}

const gchar *diversity_conn_iface_get_vendor(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	return priv->vendor;
}

const gchar *diversity_conn_iface_get_product(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	return priv->product;
}

guint diversity_conn_iface_get_signal(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	return priv->signal;
}

DiversityConnIfaceState diversity_conn_iface_get_state(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	return priv->state;
}

void diversity_conn_iface_set_policy(DiversityConnIface *cface, DiversityConnIfacePolicy policy)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);
	const gchar *p = policy_to_policy_string(policy);
	GError *error = NULL;

	if (!dbus_g_proxy_call(priv->interface, "SetPolicy", &error,
				G_TYPE_STRING, &p,
				G_TYPE_INVALID,
				G_TYPE_INVALID)) {
		g_warning("failed to set policy: %s", error->message);

		g_error_free(error);
	}
}

DiversityConnIfacePolicy diversity_conn_iface_get_policy(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	return priv->policy;
}

static guint address_from_string(const gchar *addr)
{
	const gchar *comps[4];
	const gchar *p;
	gint n;
	guint ip;

	comps[0] = addr;
	n = 1;

	p = addr;
	while (p) {
		p = strchr(p, '.');
		if (n >= 4)
			break;

		if (p) {
			p++;
			comps[n++] = p;
		}
	}

	if (p || n != 4)
		return 0;

	ip = atoi(comps[0]) | atoi(comps[1]) << 8 | atoi(comps[2]) << 16 | atoi(comps[3]) << 24;

	return ip;
}

static void ipv4_from_hash(const gchar *key, GValue *val, DiversityConnIPv4 *ipv4)
{
	const gchar *p;

	if (!G_VALUE_HOLDS_STRING(val)) {
		g_debug("unrecognized key %s", key);

		return;
	}

	p = g_value_get_string(val);

	if (strcmp(key, "Address") == 0)
		ipv4->address = address_from_string(p);
	else if (strcmp(key, "Netmask") == 0)
		ipv4->netmask = address_from_string(p);
	else if (strcmp(key, "Gateway") == 0)
		ipv4->gateway = address_from_string(p);
	else if (strcmp(key, "Method") == 0)
		ipv4->dhcp = (strcmp(p, "dhcp") == 0);
	else
		g_debug("unrecognized key %s", key);
}

const DiversityConnIPv4 *diversity_conn_iface_get_ipv4(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);
	GHashTable *hash;
	GError *error = NULL;
	GType type;

	type = dbus_g_type_get_map("GHashTable",
			G_TYPE_STRING, G_TYPE_VALUE);

	memset(priv->ipv4, 0, sizeof(*priv->ipv4));

	if (!dbus_g_proxy_call(priv->interface, "GetIPv4", &error,
				G_TYPE_INVALID,
				type, &hash,
				G_TYPE_INVALID)) {
		g_warning("failed to get ipv4: %s", error->message);

		g_error_free(error);

		return priv->ipv4;
	}

	g_hash_table_foreach(hash, (GHFunc) ipv4_from_hash, priv->ipv4);
	g_hash_table_destroy(hash);

	return priv->ipv4;
}

void diversity_conn_iface_set_network(DiversityConnIface *cface, DiversityConnNetwork *net)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);
	GHashTable *hash;
	GError *error = NULL;
	GType type;

	type = dbus_g_type_get_map("GHashTable",
			G_TYPE_STRING, G_TYPE_VALUE);

	hash = diversity_conn_network_hash(net);

	if (!dbus_g_proxy_call(priv->interface, "SetNetwork", &error,
				type, hash,
				G_TYPE_INVALID,
				G_TYPE_INVALID)) {
		g_warning("failed to set network: %s", error->message);

		g_error_free(error);
	}

	g_hash_table_destroy(hash);
}

const DiversityConnNetwork *diversity_conn_iface_get_network(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);

	return priv->active;
}

void diversity_conn_iface_scan(DiversityConnIface *cface)
{
	DiversityConnIfacePrivate *priv =
		DIVERSITY_CONN_IFACE_GET_PRIVATE(cface);
	GError *error = NULL;

	if (!dbus_g_proxy_call(priv->interface, "Scan", &error,
				G_TYPE_INVALID,
				G_TYPE_INVALID)) {
		g_warning("failed to scan: %s", error->message);

		g_error_free(error);
	}
}
