/* diversity-conn-man.c - DiversityConnMan
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <dbus/dbus-glib.h>
#include <string.h>
#include "diversity-connman.h"
#include "diversity-conn-man.h"

enum {
	INTERFACE_ADDED,
	INTERFACE_REMOVED,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_STATE,
};


#define DIVERSITY_CONN_MAN_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_CONN_MAN, DiversityConnManPrivate))
typedef struct _DiversityConnManPrivate DiversityConnManPrivate;

struct _DiversityConnManPrivate {
	DBusGConnection *connection;

	DBusGProxy *manager;
	DiversityConnManState state;

	GHashTable *ifaces;
};

static guint conn_man_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(DiversityConnMan, diversity_conn_man, G_TYPE_OBJECT);

static void diversity_conn_man_dispose(GObject *obj)
{
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(obj);

	if (priv->manager) {
		g_object_unref(priv->manager);
		priv->manager = NULL;
	}

	((GObjectClass *) diversity_conn_man_parent_class)->dispose(obj);
}

static void diversity_conn_man_finalize(GObject *obj)
{
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(obj);

	g_hash_table_destroy(priv->ifaces);
	dbus_g_connection_unref(priv->connection);

	((GObjectClass *) diversity_conn_man_parent_class)->finalize(obj);
}

static void diversity_conn_man_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	switch (prop_id) {
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_conn_man_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec)
{
	DiversityConnMan *cman = DIVERSITY_CONN_MAN(obj);
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(cman);

	switch (prop_id) {
	case PROP_STATE:
		g_value_set_int(value, priv->state);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_conn_man_class_init(DiversityConnManClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;

	o_class->dispose = diversity_conn_man_dispose;
	o_class->finalize = diversity_conn_man_finalize;
	o_class->set_property = diversity_conn_man_set_property;
	o_class->get_property = diversity_conn_man_get_property;

	conn_man_signals[INTERFACE_ADDED] =
		g_signal_new("interface-added",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityConnManClass,
					     interface_added),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__STRING,
			     G_TYPE_NONE, 1,
			     G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);

	conn_man_signals[INTERFACE_REMOVED] =
		g_signal_new("interface-removed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityConnManClass,
					     interface_removed),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__STRING,
			     G_TYPE_NONE, 1,
			     G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);

	g_object_class_install_property(o_class,
					PROP_STATE,
					g_param_spec_int(
						"state",
						"State",
						"Manager state",
						DIVERSITY_CONN_MAN_STATE_OFFLINE,
						DIVERSITY_CONN_MAN_STATE_ONLINE,
						DIVERSITY_CONN_MAN_STATE_OFFLINE,
						G_PARAM_READABLE));

	g_type_class_add_private(klass, sizeof(DiversityConnManPrivate));
}

static void diversity_conn_man_init(DiversityConnMan *cman)
{
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(cman);

	priv->state = DIVERSITY_CONN_MAN_STATE_OFFLINE;

	priv->ifaces = g_hash_table_new_full(
			g_str_hash, g_str_equal,
			g_free, NULL);
}

static DiversityConnManState state_string_to_state(const gchar *state)
{
	DiversityConnManState s;

	if (!state)
		return DIVERSITY_CONN_MAN_STATE_OFFLINE;

	if (strcmp(state, "online") == 0)
		s = DIVERSITY_CONN_MAN_STATE_ONLINE;
	else
		s = DIVERSITY_CONN_MAN_STATE_OFFLINE;

	return s;
}

static gboolean add_interface(DiversityConnMan *cman, const gchar *objpath)
{
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(cman);
	gchar *p;

	if (g_hash_table_lookup(priv->ifaces, objpath))
		return FALSE;

	p = g_strdup(objpath);
	g_hash_table_insert(priv->ifaces, p, p);

	return TRUE;
}

static gboolean remove_interface(DiversityConnMan *cman, const gchar *objpath)
{
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(cman);

	return g_hash_table_remove(priv->ifaces, objpath);
}

static void on_interface_added(DBusGProxy *proxy, const gchar *objpath, DiversityConnMan *cman)
{
	if (add_interface(cman, objpath))
		g_signal_emit(cman, conn_man_signals[INTERFACE_ADDED], 0,
				objpath);
}

static void on_interface_removed(DBusGProxy *proxy, const gchar *objpath, DiversityConnMan *cman)
{
	if (remove_interface(cman, objpath))
		g_signal_emit(cman, conn_man_signals[INTERFACE_REMOVED], 0,
				objpath);
}

static void on_state_changed(DBusGProxy *proxy, const gchar *state, DiversityConnMan *cman)
{
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(cman);
	DiversityConnManState s = state_string_to_state(state);

	if (priv->state != s) {
		priv->state = s;

		g_object_notify(G_OBJECT(cman), "state");
	}
}

static gboolean diversity_conn_man_bind(DiversityConnMan *cman)
{
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(cman);
	gchar *state;
	GPtrArray *ifaces;
	GType type;

	dbus_g_proxy_add_signal(priv->manager, "StateChanged",
				G_TYPE_STRING, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(priv->manager, "StateChanged",
				G_CALLBACK(on_state_changed), cman, NULL);

	if (dbus_g_proxy_call(priv->manager, "GetState", NULL,
				G_TYPE_INVALID,
				G_TYPE_STRING, &state,
				G_TYPE_INVALID)) {
		priv->state = state_string_to_state(state);
		g_free(state);
	}

	dbus_g_proxy_add_signal(priv->manager, "InterfaceAdded",
				DBUS_TYPE_G_OBJECT_PATH, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(priv->manager, "InterfaceAdded",
				G_CALLBACK(on_interface_added), cman, NULL);

	dbus_g_proxy_add_signal(priv->manager, "InterfaceRemoved",
				DBUS_TYPE_G_OBJECT_PATH, G_TYPE_INVALID);

	dbus_g_proxy_connect_signal(priv->manager, "InterfaceRemoved",
				G_CALLBACK(on_interface_removed), cman, NULL);

	type = dbus_g_type_get_collection("GPtrArray",
			DBUS_TYPE_G_OBJECT_PATH);

	if (dbus_g_proxy_call(priv->manager, "ListInterfaces", NULL,
				G_TYPE_INVALID,
				type, &ifaces,
				G_TYPE_INVALID)) {
		gint i;

		for (i = 0; i < ifaces->len; i++) {
			gchar *objpath = g_ptr_array_index(ifaces, i);

			add_interface(cman, objpath);

			g_free(objpath);
		}

		g_ptr_array_free(ifaces, TRUE);
	}

	return TRUE;
}

DiversityConnMan *diversity_conn_man_new(GError **error)
{
	DiversityConnMan *cman;
	DiversityConnManPrivate *priv;
	
	cman = g_object_new(DIVERSITY_TYPE_CONN_MAN, NULL);
	priv = DIVERSITY_CONN_MAN_GET_PRIVATE(cman);

	priv->connection = dbus_g_bus_get(CONNMAN_DBUS_BUS, error);
	if (!priv->connection) {
		g_object_unref(cman);

		return NULL;
	}

	priv->manager =
		dbus_g_proxy_new_for_name(priv->connection,
					  CONNMAN_DBUS_SERVICE,
					  CONNMAN_MANAGER_DBUS_PATH,
					  CONNMAN_MANAGER_DBUS_INTERFACE);
	if (!priv->manager) {
		g_set_error(error, 0, 0, "failed to create proxy");

		g_object_unref(cman);

		return NULL;
	}

	if (!diversity_conn_man_bind(cman)) {
		g_set_error(error, 0, 0, "failed to bind manager");
		g_object_unref(cman);

		return NULL;
	}

	return cman;
}

DiversityConnManState diversity_conn_man_get_state(DiversityConnMan *cman)
{
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(cman);

	return priv->state;
}

static void list_interfaces(gpointer key, gpointer val, gpointer data)
{
	GSList **iface = data;

	*iface = g_slist_prepend(*iface, val);
}

GSList *diversity_conn_man_list_interfaces(DiversityConnMan *cman)
{
	DiversityConnManPrivate *priv = DIVERSITY_CONN_MAN_GET_PRIVATE(cman);
	GSList *ifaces = NULL;

	g_hash_table_foreach(priv->ifaces, list_interfaces, &ifaces);

	return ifaces;
}
