/* diversity-conn-nm.c - DiversityConnNm
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <diversity-world.h>
#include <diversity-utils.h>
#include <diversity-equipment.h>
#include <diversity-net.h>
#include <diversity-ap.h>
#include "diversity-conn-nm.h"
#include "diversity-conn-man.h"
#include "diversity-conn-iface.h"

enum {
	PROP_0,
	PROP_STATE,
	PROP_INTERFACE,
	PROP_IP4_ADDRESS,
	PROP_ACTIVE_AP,
};


#define DIVERSITY_CONN_NM_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_CONN_NM, DiversityConnNmPrivate))
typedef struct _DiversityConnNmPrivate DiversityConnNmPrivate;

struct _DiversityConnNmPrivate {
	DiversityWorld *world;

	DiversityConnMan *cman;
	DiversityConnIface *cface;

	DiversityAp *active_ap;
};

static void diversity_conn_nm_iface_init(gpointer g_iface, gpointer iface_data);

G_DEFINE_TYPE_EXTENDED(DiversityConnNm, diversity_conn_nm, DIVERSITY_TYPE_EQUIPMENT, 0,
		G_IMPLEMENT_INTERFACE(DIVERSITY_TYPE_NET, diversity_conn_nm_iface_init));

static void diversity_conn_nm_dispose(GObject *obj)
{
	DiversityConnNmPrivate *priv = DIVERSITY_CONN_NM_GET_PRIVATE(obj);

	if (priv->active_ap) {
		g_object_unref(priv->active_ap);

		priv->active_ap = NULL;
	}

	if (priv->cface) {
		g_object_unref(priv->cface);

		priv->cface = NULL;
	}

	if (priv->cman) {
		g_object_unref(priv->cman);

		priv->cman = NULL;
	}

	if (priv->world) {
		g_object_unref(priv->world);

		priv->world = NULL;
	}

	((GObjectClass *) diversity_conn_nm_parent_class)->dispose(obj);
}

static void diversity_conn_nm_finalize(GObject *obj)
{
	((GObjectClass *) diversity_conn_nm_parent_class)->finalize(obj);
}

static void diversity_conn_nm_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	switch (prop_id) {
	case PROP_INTERFACE:
	case PROP_STATE:
	case PROP_IP4_ADDRESS:
	case PROP_ACTIVE_AP:
		/* fall through */
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static DiversityNetState map_state(DiversityConnIfaceState state)
{
	DiversityNetState s;

	switch (state) {
	case DIVERSITY_CONN_IFACE_STATE_OFF:
	case DIVERSITY_CONN_IFACE_STATE_SHUTDOWN:
		s = DIVERSITY_NET_STATE_DOWN;
		break;
	case DIVERSITY_CONN_IFACE_STATE_ENABLED:
	case DIVERSITY_CONN_IFACE_STATE_SCANNING:
		s = DIVERSITY_NET_STATE_DISCONNECTED;
		break;
	case DIVERSITY_CONN_IFACE_STATE_CONNECT:
		s = DIVERSITY_NET_STATE_PREPARE;
		break;
	case DIVERSITY_CONN_IFACE_STATE_CONNECTED:
	case DIVERSITY_CONN_IFACE_STATE_CARRIER:
		s = DIVERSITY_NET_STATE_CONFIG;
		break;
	case DIVERSITY_CONN_IFACE_STATE_CONFIGURE:
		s = DIVERSITY_NET_STATE_IP_CONFIG;
		break;
	case DIVERSITY_CONN_IFACE_STATE_READY:
		s = DIVERSITY_NET_STATE_ACTIVATED;
		break;
	default:
	case DIVERSITY_CONN_IFACE_STATE_UNKNOWN:
		s = DIVERSITY_NET_STATE_UNKNOWN;
		break;
	}

	return s;
}

static void diversity_conn_nm_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec)
{
	DiversityConnNm *cnm = DIVERSITY_CONN_NM(obj);
	DiversityConnNmPrivate *priv = DIVERSITY_CONN_NM_GET_PRIVATE(cnm);

	switch (prop_id) {
	case PROP_INTERFACE:
		g_value_set_static_string(value, "unk");
		break;
	case PROP_STATE:
		{
			DiversityConnIfaceState s;
			DiversityNetState state;

			s = diversity_conn_iface_get_state(priv->cface);
			state = map_state(s);
			g_value_set_uint(value, state);
		}
		break;
	case PROP_IP4_ADDRESS:
		{
			const DiversityConnIPv4 *ipv4;

			ipv4 = diversity_conn_iface_get_ipv4(priv->cface);
			g_value_set_uint(value, ipv4->address);
		}
		break;
	case PROP_ACTIVE_AP:
		g_value_set_object(value, priv->active_ap);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static gboolean diversity_conn_nm_equip(DiversityEquipment *eqp)
{
	return TRUE;
}

static void diversity_conn_nm_unequip(DiversityEquipment *eqp)
{
}

static gboolean diversity_conn_nm_set_config(DiversityEquipment *eqp, const gchar *key, const GValue *val)
{
	return FALSE;
}

static gboolean diversity_conn_nm_get_config(DiversityEquipment *eqp, const gchar *key, GValue *val)
{
	return FALSE;
}

static void diversity_conn_nm_class_init(DiversityConnNmClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;
	DiversityEquipmentClass *eqp_class = (DiversityEquipmentClass *) klass;

	o_class->dispose = diversity_conn_nm_dispose;
	o_class->finalize = diversity_conn_nm_finalize;
	o_class->set_property = diversity_conn_nm_set_property;
	o_class->get_property = diversity_conn_nm_get_property;

	eqp_class->equip = diversity_conn_nm_equip;
	eqp_class->unequip = diversity_conn_nm_unequip;

	eqp_class->set_config = diversity_conn_nm_set_config;
	eqp_class->get_config = diversity_conn_nm_get_config;

	g_object_class_override_property(o_class,
			PROP_INTERFACE, "interface");
	g_object_class_override_property(o_class,
			PROP_STATE, "state");
	g_object_class_override_property(o_class,
			PROP_IP4_ADDRESS, "ip4-address");
	g_object_class_override_property(o_class,
			PROP_ACTIVE_AP, "active-ap");

	g_type_class_add_private(klass, sizeof(DiversityConnNmPrivate));
}

static void diversity_conn_nm_init(DiversityConnNm *cnm)
{
}

static void position_ap(DiversityAp *ap, DiversityObject *self)
{
	DiversityObjectAccuracy accuracy;
	double lon, lat, len;

	accuracy = diversity_object_get_accuracy(DIVERSITY_OBJECT(ap));
	if (accuracy != DIVERSITY_OBJECT_ACCURACY_NONE)
		return;

	accuracy = diversity_object_get_accuracy(self);
	if (accuracy == DIVERSITY_OBJECT_ACCURACY_NONE)
		return;
	else if (accuracy >= DIVERSITY_OBJECT_ACCURACY_DIRECT)
		accuracy = DIVERSITY_OBJECT_ACCURACY_DERIVED;

	diversity_object_set_accuracy(DIVERSITY_OBJECT(ap), accuracy);

	diversity_object_geometry_get(self, &lon, &lat, NULL, NULL);
	len = DIVERSITY_UNITS(200);
	lon -= len / 2;
	lat -= len / 2;

	diversity_object_geometry_set(DIVERSITY_OBJECT(ap),
			lon, lat, len, len);
}

static void update_ap(DiversityAp *ap, const DiversityConnNetwork *net)
{
	diversity_ap_set_ssid(ap, net->essid);
	diversity_ap_set_hw_address(ap, net->bssid);
	diversity_ap_set_flags(ap, (net->security) ?
			DIVERSITY_AP_FLAG_PRIVACY :
			0);
	diversity_ap_set_strength(ap, net->signal);
	diversity_ap_set_in_service(ap, TRUE);
}

static void add_ap(DiversityWorld *world, const DiversityConnNetwork *net)
{
	DiversityAp *ap;
	DiversityObject *self;

	ap = diversity_ap_new();
	DIVERSITY_OBJECT(ap)->id = g_strdup(net->bssid);
	update_ap(ap, net);

	self = (DiversityObject *) diversity_world_get_self(world);
	position_ap(ap, self);

	diversity_world_add_object(world, DIVERSITY_OBJECT(ap));

	g_debug("commit ap: %c%s (%s)",
			diversity_ap_get_flags(ap) ? '*' : ' ',
			diversity_ap_get_ssid(ap),
			diversity_ap_get_hw_address(ap));
}

static void on_ap_added(DiversityConnIface *cface, DiversityConnNetwork *net, DiversityConnNm *cnm)
{
	DiversityConnNmPrivate *priv = DIVERSITY_CONN_NM_GET_PRIVATE(cnm);
	DiversityObject *ap;

	if (!net->bssid)
		return;

	ap = diversity_world_lookup_object(priv->world,
			DIVERSITY_OBJECT_TYPE_AP, net->bssid);
	if (ap) {
		DiversityObject *self;

		self = (DiversityObject *)
			diversity_world_get_self(priv->world);

		update_ap(DIVERSITY_AP(ap), net);
		position_ap(DIVERSITY_AP(ap), self);
	} else {
		add_ap(priv->world, net);
	}
}

static void on_ap_removed(DiversityConnIface *cface, DiversityConnNetwork *net, DiversityConnNm *cnm)
{
	DiversityConnNmPrivate *priv = DIVERSITY_CONN_NM_GET_PRIVATE(cnm);
	DiversityAp *ap;

	if (!net->bssid)
		return;

	ap = (DiversityAp *) diversity_world_lookup_object(priv->world,
			DIVERSITY_OBJECT_TYPE_AP, net->bssid);
	if (ap)
		diversity_ap_set_in_service(ap, FALSE);
}

static void on_ap_changed(DiversityConnIface *cface, DiversityConnNetwork *net, DiversityConnNm *cnm)
{
	DiversityConnNmPrivate *priv = DIVERSITY_CONN_NM_GET_PRIVATE(cnm);
	DiversityAp *ap;
	DiversityObject *self;

	if (!net->bssid)
		return;

	ap = (DiversityAp *) diversity_world_lookup_object(priv->world,
			DIVERSITY_OBJECT_TYPE_AP, net->bssid);
	if (!ap) {
		g_debug("update non-existing ap: %s", net->essid);

		add_ap(priv->world, net);

		return;
	}

	update_ap(ap, net);

	self = (DiversityObject *) diversity_world_get_self(priv->world);
	position_ap(ap, self);
}

static void on_state_changed(DiversityConnIface *cface, GParamSpec *pspec, DiversityConnNm *cnm)
{
	DiversityConnIfaceState state = diversity_conn_iface_get_state(cface);

	g_debug("state changed to %d", state);

	g_signal_emit_by_name(cnm, "state-changed", map_state(state));
}

DiversityConnNm *diversity_conn_nm_new(DiversityWorld *world, GError **error)
{
	DiversityConnNm *cnm;
	DiversityConnNmPrivate *priv;
	GSList *cfaces, *tmp_list;
	const DiversityConnNetwork *net;

	cnm = g_object_new(DIVERSITY_TYPE_CONN_NM,
			    "name", "connman",
			    NULL);

	priv = DIVERSITY_CONN_NM_GET_PRIVATE(cnm);

	priv->cman = diversity_conn_man_new(error);
	if (!priv->cman) {
		g_object_unref(cnm);

		return NULL;
	}

	cfaces = diversity_conn_man_list_interfaces(priv->cman);

	tmp_list = cfaces;
	while (tmp_list) {
		DiversityConnIface *cface;
		const gchar *objpath = tmp_list->data;

		cface = diversity_conn_iface_new(objpath, NULL);
		if (cface) {
			if (diversity_conn_iface_get_iface_type(cface) ==
					DIVERSITY_CONN_IFACE_TYPE_80211) {
				priv->cface = cface;

				break;
			}

			g_object_unref(cface);
		}

		g_debug("ignore interface %s", objpath);

		tmp_list = tmp_list->next;
	}

	if (cfaces)
		g_slist_free(cfaces);

	if (!priv->cface) {
		g_set_error(error, 0, 0, "no supported interface");

		g_object_unref(cnm);

		return NULL;
	}

	g_signal_connect(priv->cface, "ap-added",
			G_CALLBACK(on_ap_added), cnm);
	g_signal_connect(priv->cface, "ap-removed",
			G_CALLBACK(on_ap_removed), cnm);
	g_signal_connect(priv->cface, "ap-changed",
			G_CALLBACK(on_ap_changed), cnm);
	g_signal_connect(priv->cface, "notify::state",
			G_CALLBACK(on_state_changed), cnm);

	priv->world = world;
	g_object_ref(priv->world);

	net = diversity_conn_iface_get_network(priv->cface);
	if (net) {
		DiversityAp *ap;

		ap = diversity_ap_new();
		g_object_ref_sink(ap);

		update_ap(ap, net);

		priv->active_ap = ap;
	}

	return cnm;
}

/* DiversityNet method implementations */

static gboolean diversity_conn_nm_connect_ap(DiversityNet *net, DiversityAp *ap, GError **error)
{
	DiversityConnNmPrivate *priv = DIVERSITY_CONN_NM_GET_PRIVATE(net);
	DiversityConnNetwork network = { 0 };

	network.essid = (gchar *) diversity_ap_get_ssid(ap);
	network.psk = "";

	if (priv->active_ap)
		g_object_unref(priv->active_ap);
	priv->active_ap = ap;
	g_object_ref(ap);

	diversity_conn_iface_set_network(priv->cface, &network);

	return TRUE;
}

static gboolean diversity_conn_nm_connect_wired(DiversityNet *net, GError **error)
{
	g_set_error(error, 0, 0, "NOT IMPLEMENTED");

	return FALSE;
}

static void diversity_conn_nm_iface_init(gpointer g_iface, gpointer iface_data)
{
	DiversityNetIface *iface = (DiversityNetIface *) g_iface;

	iface->connect_ap = diversity_conn_nm_connect_ap;
	iface->connect_wired = diversity_conn_nm_connect_wired;
}
