/* diversity-iwlib-backend.c -
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <diversity-iwlib-backend.h>
#include <diversity-utils.h>
#include <errno.h>
#include <string.h> /* for memset */
#include "iwlib.h"
#include "diversity-nm.h"

static struct {
	DiversityWorld *world;
	DiversityNm *nm;
} iwlib;

static void position_ap(DiversityAp *ap)
{
	gdouble lon, lat, len;

	diversity_world_get_position(iwlib.world, &lon, &lat);

	len = DIVERSITY_UNITS(200);
	lon -= len / 2;
	lat -= len / 2;

	diversity_object_geometry_set(DIVERSITY_OBJECT(ap),
			lon, lat, len, len);
	diversity_object_set_accuracy(DIVERSITY_OBJECT(ap),
			DIVERSITY_OBJECT_ACCURACY_DERIVED);
}

static void on_ap_added(DiversityNm *nm, GSList *aps, gpointer user_data)
{
	GSList *tmp_list;

	tmp_list = aps;
	while (tmp_list) {
		DiversityAp *ap = tmp_list->data;
		DiversityObject *old;
		const gchar *id;

		id = diversity_ap_get_hw_address(ap);
		old = diversity_world_lookup_object(iwlib.world,
				DIVERSITY_OBJECT_TYPE_AP, id);
		if (old) {
			ap = DIVERSITY_AP(old);
		} else {
			DIVERSITY_OBJECT(ap)->id = g_strdup(id);
			position_ap(ap);

			diversity_world_add_object(iwlib.world, DIVERSITY_OBJECT(ap));

			g_debug("commit ap for %s: %c%s (%s)",
					diversity_nm_get_interface(iwlib.nm),
					diversity_ap_get_flags(ap) ? '*' : ' ',
					diversity_ap_get_ssid(ap),
					id);
		}

		diversity_ap_set_in_service(ap, TRUE);

		tmp_list = tmp_list->next;
	}
}

static void on_ap_removed(DiversityNm *nm, GSList *aps, gpointer user_data)
{
	GSList *tmp_list;

	tmp_list = aps;
	while (tmp_list) {
		DiversityAp *ap = tmp_list->data;
		const gchar *id;

		id = diversity_ap_get_hw_address(ap);
		ap = (DiversityAp *)
			diversity_world_lookup_object(iwlib.world,
				DIVERSITY_OBJECT_TYPE_AP, id);
		if (ap)
			diversity_ap_set_in_service(ap, FALSE);

		tmp_list = tmp_list->next;
	}
}

static void on_accuracy_changed(DiversityBard *self)
{
	DiversityObjectAccuracy accuracy;

	accuracy = diversity_object_get_accuracy(DIVERSITY_OBJECT(self));

	if (accuracy == DIVERSITY_OBJECT_ACCURACY_NONE)
		diversity_nm_set_scan_interval(iwlib.nm, 0);
	else
		diversity_nm_set_scan_interval(iwlib.nm, 15);
}

int iwlib_init(DiversityWorld *world, GError **error)
{
	DiversityBard *self;

	iwlib.world = world;
	g_object_ref(world);

	iwlib.nm = diversity_nm_new();
	diversity_world_net_set(iwlib.world, DIVERSITY_NET(iwlib.nm));

	g_signal_connect(iwlib.nm, "ap-added",
		G_CALLBACK(on_ap_added), NULL);
	g_signal_connect(iwlib.nm, "ap-removed",
		G_CALLBACK(on_ap_removed), NULL);

	self = diversity_world_get_self(iwlib.world);
	g_signal_connect(self, "notify::accuracy",
		G_CALLBACK(on_accuracy_changed), NULL);

	on_accuracy_changed(self);

	diversity_bard_add_equipment(self,
			DIVERSITY_EQUIPMENT(iwlib.nm));

	return TRUE;
}

void iwlib_fini(void)
{
	if (iwlib.nm) {
		g_object_unref(iwlib.nm);
		iwlib.nm = NULL;
	}

	if (iwlib.world) {
		g_object_unref(G_OBJECT(iwlib.world));
		iwlib.world = NULL;
	}
}
