/* diversity-iwlib.c - DiversityIwlib
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@opeiwliboko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include "diversity-iwlib.h"
#include "iwlib.h"

enum {
	SCANNED_AP,
	STATE_CHANGED,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_INTERFACE,
};


#define DIVERSITY_IWLIB_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_IWLIB, DiversityIwlibPrivate))
typedef struct _DiversityIwlibPrivate DiversityIwlibPrivate;

struct _DiversityIwlibPrivate {
	gint fd;

	gchar *iface;
	guint state;
	guint assoc_id;

	guint scan_delay;
	guint scan_interval;
	guint scan_id;
	guint scan_count;
	GByteArray *array;
	guint max_size;

	gboolean scanning;
	gboolean has_range;
	struct iw_range range;

	GSList *aps;

	DiversityAp *assoc_ap;
};

static guint iwlib_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(DiversityIwlib, diversity_iwlib, G_TYPE_OBJECT);

static void diversity_iwlib_dispose(GObject *obj)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(obj);

	if (priv->assoc_ap) {
		g_object_unref(priv->assoc_ap);

		priv->assoc_ap = NULL;
	}

	((GObjectClass *) diversity_iwlib_parent_class)->dispose(obj);
}

static void diversity_iwlib_finalize(GObject *obj)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(obj);

	if (priv->fd >= 0)
		iw_sockets_close(priv->fd);
	g_free(priv->iface);
	g_byte_array_free(priv->array, TRUE);

	if (priv->assoc_id)
		g_source_remove(priv->assoc_id);

	((GObjectClass *) diversity_iwlib_parent_class)->finalize(obj);
}

static void diversity_iwlib_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	DiversityIwlib *iwlib = DIVERSITY_IWLIB(obj);
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	switch (prop_id) {
	case PROP_INTERFACE:
		g_free(priv->iface);
		priv->iface = g_value_dup_string(value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_iwlib_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec)
{
	DiversityIwlib *iwlib = DIVERSITY_IWLIB(obj);
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	switch (prop_id) {
	case PROP_INTERFACE:
		g_value_set_string(value, priv->iface);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_iwlib_class_init(DiversityIwlibClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;

	o_class->dispose = diversity_iwlib_dispose;
	o_class->finalize = diversity_iwlib_finalize;
	o_class->set_property = diversity_iwlib_set_property;
	o_class->get_property = diversity_iwlib_get_property;

	iwlib_signals[SCANNED_AP] =
		g_signal_new("scanned-ap",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityIwlibClass, scanned_ap),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	iwlib_signals[STATE_CHANGED] =
		g_signal_new("state-changed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityIwlibClass, state_changed),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__UINT,
			     G_TYPE_NONE, 1,
			     G_TYPE_UINT);

	g_object_class_install_property(o_class,
					PROP_INTERFACE,
					g_param_spec_string("interface",
						 	    "Interface",
							    "Interface name",
							    "",
							    G_PARAM_READWRITE |
							    G_PARAM_CONSTRUCT_ONLY));

	g_type_class_add_private(klass, sizeof(DiversityIwlibPrivate));
}

static void diversity_iwlib_init(DiversityIwlib *iwlib)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	priv->iface = NULL;
	priv->state = DIVERSITY_IWLIB_STATE_DISCONNECTED;

	priv->scan_delay = 250;
	priv->scan_interval = 100;

	priv->array = g_byte_array_new();
	priv->max_size = 0x100000;
}

static void diversity_iwlib_set_state(DiversityIwlib *iwlib, guint state)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	if (priv->state != state) {
		priv->state = state;

		g_signal_emit(iwlib, iwlib_signals[STATE_CHANGED], 0,
				priv->state);
	}
}

static gboolean assoc_monitor(gpointer data)
{
	DiversityIwlib *iwlib = data;
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);
	iwstats stats;

	if (!priv->assoc_ap) {
		diversity_iwlib_set_state(iwlib,
				DIVERSITY_IWLIB_STATE_DISCONNECTED);

		return TRUE;
	}

	if (iw_get_stats(priv->fd, priv->iface, &stats,
				&priv->range, priv->has_range) < 0 ||
	    !stats.qual.qual) {
		diversity_iwlib_set_state(iwlib,
				DIVERSITY_IWLIB_STATE_ASSOCIATING);

		return TRUE;
	}

	/*
	g_debug("associated with %s (qual %d)",
			diversity_ap_get_ssid(priv->assoc_ap),
			stats.qual.qual);
			*/

	diversity_iwlib_set_state(iwlib, DIVERSITY_IWLIB_STATE_ASSOCIATED);

	return TRUE;
}

DiversityIwlib *diversity_iwlib_new(const gchar *iface)
{
	DiversityIwlib *iwlib;
	DiversityIwlibPrivate *priv;
	wireless_config info;

	iwlib = g_object_new(DIVERSITY_TYPE_IWLIB,
			    "interface", iface,
			    NULL);

	priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	priv->fd = iw_sockets_open();
	if (priv->fd < 0) {
		g_warning("failed to open socket");
		g_object_unref(iwlib);

		return NULL;
	}

	if (iw_get_range_info(priv->fd, priv->iface, &priv->range) < 0 ||
	    priv->range.we_version_compiled < 14) {
		g_warning("interface %s is not supported", iface);
		g_object_unref(iwlib);

		return NULL;
	}

	priv->assoc_id = g_timeout_add_seconds(15, assoc_monitor, iwlib);

	if (iw_get_basic_config(priv->fd, priv->iface, &info) >= 0 &&
	    info.has_essid) {
		struct iwreq wrq;

		if (iw_get_ext(priv->fd, priv->iface, SIOCGIWAP, &wrq) >= 0) {
			DiversityAp *ap;
			gchar *hw_addr;
			gchar buf[128];

			hw_addr = iw_saether_ntop(&wrq.u.ap_addr, buf);

			ap = diversity_ap_new();
			diversity_ap_set_hw_address(ap, hw_addr);
			diversity_ap_set_ssid(ap, info.essid);

			priv->assoc_ap = ap;
			g_object_ref_sink(ap);

			assoc_monitor(iwlib);
		}
	}

	return iwlib;
}

guint diversity_iwlib_get_state(DiversityIwlib *iwlib)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	return priv->state;
}

gboolean diversity_iwlib_request_associate(DiversityIwlib *iwlib, DiversityAp *ap)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);
	const gchar *essid = NULL;
	struct iwreq wrq;
	gsize len;

	if (ap)
		essid = diversity_ap_get_ssid(ap);

	if (essid) {
		len = strlen(essid);

		if (len > IW_ESSID_MAX_SIZE)
			return FALSE;

		wrq.u.essid.flags = 1;

	} else {
		essid = "";
		len = 0;
		wrq.u.essid.flags = 0;

	}

	wrq.u.essid.pointer = (caddr_t) essid;
	wrq.u.essid.length = len;

	if (iw_get_kernel_we_version() < 21)
		wrq.u.essid.length++;

	if (iw_set_ext(priv->fd, priv->iface, SIOCSIWESSID, &wrq) < 0) {
		g_warning("ioctl SIOCSIWESSID failed");

		return FALSE;
	}

	if (priv->assoc_ap != ap) {
		if (priv->assoc_ap)
			g_object_unref(priv->assoc_ap);
		priv->assoc_ap = ap;
		if (ap)
			g_object_ref(ap);
	}

	assoc_monitor(iwlib);

	return TRUE;
}

gboolean diversity_iwlib_cancel_associate(DiversityIwlib *iwlib)
{
	return diversity_iwlib_request_associate(iwlib, NULL);
}

DiversityAp *diversity_iwlib_get_associate(DiversityIwlib *iwlib)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	return priv->assoc_ap;
}

static void scan_parse_event(DiversityIwlib *iwlib, struct iw_event *e)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);
	gchar buf[128];
	DiversityAp *ap = NULL;

	/* dangling events */
	if (e->cmd != SIOCGIWAP && !priv->aps)
		return;

	if (priv->aps)
		ap = priv->aps->data;

	switch (e->cmd) {
	case SIOCGIWAP:
		{
			gchar *hw_addr;

			hw_addr = iw_saether_ntop(&e->u.ap_addr, buf);

			ap = diversity_ap_new();
			g_object_ref_sink(ap);

			diversity_ap_set_hw_address(ap, hw_addr);

			priv->aps = g_slist_prepend(priv->aps, ap);
		}
		break;
	case SIOCGIWESSID:
		{
			gchar essid[IW_ESSID_MAX_SIZE + 1];
			gint index;

			if (e->u.essid.pointer && e->u.essid.length &&
					e->u.essid.flags) {
				memcpy(essid, e->u.essid.pointer,
						e->u.essid.length);
				essid[e->u.essid.length] = '\0';
			} else {
				*essid = '\0';
			}

			index = e->u.essid.flags & IW_ENCODE_INDEX;

			diversity_ap_set_ssid(ap, essid);
		}
		break;
	case SIOCGIWENCODE:
		{
			guint32 flags;

			flags = (e->u.data.flags & IW_ENCODE_DISABLED)
				? 0
				: DIVERSITY_AP_FLAG_PRIVACY;

			diversity_ap_set_flags(ap, flags);
		}
		break;
	case IWEVQUAL:
		if (priv->has_range &&
		    !(e->u.qual.updated & IW_QUAL_QUAL_INVALID)) {
			gint8 str = 0;

			guint8 updated = e->u.qual.updated;

			if ((updated & (IW_QUAL_DBM | IW_QUAL_RCPI)) ||
			    e->u.qual.level != 0)
				str = e->u.qual.qual * 100 /
					priv->range.max_qual.qual;

			diversity_ap_set_strength(ap, str);
		}
		break;
	case SIOCGIWFREQ:
	case SIOCGIWMODE:
	case SIOCGIWRATE:
	case SIOCGIWMODUL:
	case SIOCGIWNWID:
	case SIOCGIWNAME:
	case IWEVCUSTOM:
	default:
		break;
	}
}

static void scan_parse(DiversityIwlib *iwlib, guchar *data, guint len)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);
	struct stream_descr stream;
	struct iw_event iwe;
	gint ret;

	if (!len) {
		g_signal_emit(iwlib, iwlib_signals[SCANNED_AP], 0, NULL);

		return;
	}

	iw_init_event_stream(&stream, (gchar *) data, len);
	while (1) {
		ret = iw_extract_event_stream(&stream, &iwe,
				priv->range.we_version_compiled);

		if (ret <= 0)
			break;

		scan_parse_event(iwlib, &iwe);
	}

	g_signal_emit(iwlib, iwlib_signals[SCANNED_AP], 0,
			priv->aps);

	if (priv->aps) {
		g_slist_foreach(priv->aps,
				(GFunc) g_object_unref, NULL);
		g_slist_free(priv->aps);

		priv->aps = NULL;
	}
}

static gboolean scan_poll(gpointer data)
{
	DiversityIwlib *iwlib = data;
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);
	struct iwreq wrq;
	guint len = IW_SCAN_MAX_DATA;

	while (len < priv->max_size) {
		g_byte_array_set_size(priv->array, len);

		wrq.u.data.pointer = priv->array->data;
		wrq.u.data.length = len;
		wrq.u.data.flags = 0;

		if (iw_get_ext(priv->fd, priv->iface, SIOCGIWSCAN, &wrq) >= 0)
			break;

		if (errno == EAGAIN) {
			if (priv->scan_count) {
				priv->scan_count++;

				return TRUE;
			}

			priv->scan_count++;

			priv->scan_id =
				g_timeout_add(priv->scan_interval,
						scan_poll, iwlib);

			return FALSE;
		} else if (errno == E2BIG &&
			   priv->range.we_version_compiled > 16) {
			if (wrq.u.data.length > len)
				len = wrq.u.data.length;
			else
				len *= 2;
		} else {
			len = priv->max_size;

			break;
		}
	}

	if (len < priv->max_size)
		scan_parse(iwlib, priv->array->data, wrq.u.data.length);
	else
		scan_parse(iwlib, NULL, 0);

	priv->scanning = FALSE;
	priv->scan_id = 0;

	return FALSE;
}

gboolean diversity_iwlib_request_scan(DiversityIwlib *iwlib)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);
	struct iwreq wrq;
	
	if (priv->scanning)
		diversity_iwlib_cancel_scan(iwlib);

	priv->scanning = TRUE;
	priv->scan_id = 0;
	priv->scan_count = 0;

	wrq.u.data.pointer = NULL;
	wrq.u.data.flags = 0;
	wrq.u.data.length = 0;

	if (iw_set_ext(priv->fd, priv->iface, SIOCSIWSCAN, &wrq) < 0) {
		if (errno == EPERM) {
			/* poll the result immediately */
			scan_poll(iwlib);

			return TRUE;
		} else {
			priv->scanning = FALSE;

			return FALSE;
		}
	}

	priv->scan_id = g_timeout_add(priv->scan_delay,
			scan_poll, iwlib);

	return TRUE;
}

void diversity_iwlib_cancel_scan(DiversityIwlib *iwlib)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	priv->scanning = FALSE;

	if (priv->scan_id) {
		g_source_remove(priv->scan_id);
		priv->scan_id = 0;
	}
}

gboolean diversity_iwlib_is_scanning(DiversityIwlib *iwlib)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	return priv->scanning;
}

const gchar *diversity_iwlib_get_interface(DiversityIwlib *iwlib)
{
	DiversityIwlibPrivate *priv = DIVERSITY_IWLIB_GET_PRIVATE(iwlib);

	return priv->iface;
}
