/* diversity-nm.c - DiversityNm
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <diversity-equipment.h>
#include <diversity-net.h>
#include "diversity-nm.h"
#include "diversity-iwlib.h"

enum {
	AP_ADDED,
	AP_REMOVED,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_INTERFACE,
	PROP_STATE,
	PROP_IP4_ADDRESS,
	PROP_ACTIVE_AP,

	PROP_SCAN_INTERVAL,
};


#define DIVERSITY_NM_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_NM, DiversityNmPrivate))
typedef struct _DiversityNmPrivate DiversityNmPrivate;

struct _DiversityNmPrivate {
	gchar *iface;
	guint state;
	guint ip4_address;
	DiversityAp *active_ap;

	DiversityIwlib *iwlib;

	guint scan_interval;
	guint scan_id;
	GSList *aps;

	guint dhcp;
};

static void diversity_nm_iface_init(gpointer g_iface, gpointer iface_data);

static guint nm_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE_EXTENDED(DiversityNm, diversity_nm, DIVERSITY_TYPE_EQUIPMENT, 0,
		G_IMPLEMENT_INTERFACE(DIVERSITY_TYPE_NET, diversity_nm_iface_init));

static void diversity_nm_dispose(GObject *obj)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(obj);

	if (priv->iwlib) {
		g_object_unref(priv->iwlib);

		priv->iwlib = NULL;
	}

	if (priv->aps) {
		g_slist_foreach(priv->aps, (GFunc) g_object_unref, NULL);
		g_slist_free(priv->aps);
		priv->aps = NULL;
	}

	if (priv->active_ap) {
		g_object_unref(priv->active_ap);
		priv->active_ap = NULL;
	}

	((GObjectClass *) diversity_nm_parent_class)->dispose(obj);
}

static void diversity_nm_finalize(GObject *obj)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(obj);

	g_free(priv->iface);
	if (priv->dhcp)
		g_source_remove(priv->dhcp);

	((GObjectClass *) diversity_nm_parent_class)->finalize(obj);
}

static void diversity_nm_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	DiversityNm *nm = DIVERSITY_NM(obj);

	switch (prop_id) {
	case PROP_INTERFACE:
		diversity_nm_set_interface(nm, g_value_get_string(value));
		break;
	case PROP_STATE:
		diversity_nm_set_state(nm, g_value_get_uint(value));
		break;
	case PROP_IP4_ADDRESS:
		diversity_nm_set_ip4_address(nm, g_value_get_uint(value));
		break;
	case PROP_SCAN_INTERVAL:
		diversity_nm_set_scan_interval(nm, g_value_get_uint(value));
		break;
	case PROP_ACTIVE_AP:
		/* fall through */
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_nm_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec)
{
	DiversityNm *nm = DIVERSITY_NM(obj);
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	switch (prop_id) {
	case PROP_INTERFACE:
		g_value_set_string(value, priv->iface);
		break;
	case PROP_STATE:
		g_value_set_uint(value, priv->state);
		break;
	case PROP_IP4_ADDRESS:
		g_value_set_uint(value, priv->ip4_address);
		break;
	case PROP_SCAN_INTERVAL:
		g_value_set_uint(value, priv->scan_interval);
		break;
	case PROP_ACTIVE_AP:
		g_value_set_object(value, priv->active_ap);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static gboolean diversity_nm_equip(DiversityEquipment *eqp)
{
	return TRUE;
}

static void diversity_nm_unequip(DiversityEquipment *eqp)
{
}

static gboolean diversity_nm_set_config(DiversityEquipment *eqp, const gchar *key, const GValue *val)
{
	return FALSE;
}

static gboolean diversity_nm_get_config(DiversityEquipment *eqp, const gchar *key, GValue *val)
{
	return FALSE;
}

static void diversity_nm_class_init(DiversityNmClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;
	DiversityEquipmentClass *eqp_class = (DiversityEquipmentClass *) klass;

	o_class->dispose = diversity_nm_dispose;
	o_class->finalize = diversity_nm_finalize;
	o_class->set_property = diversity_nm_set_property;
	o_class->get_property = diversity_nm_get_property;

	eqp_class->equip = diversity_nm_equip;
	eqp_class->unequip = diversity_nm_unequip;

	eqp_class->set_config = diversity_nm_set_config;
	eqp_class->get_config = diversity_nm_get_config;

	nm_signals[AP_ADDED] =
		g_signal_new("ap-added",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityNmClass, ap_added),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	nm_signals[AP_REMOVED] =
		g_signal_new("ap-removed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityNmClass, ap_added),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__POINTER,
			     G_TYPE_NONE, 1,
			     G_TYPE_POINTER);

	g_object_class_override_property(o_class,
			PROP_INTERFACE, "interface");
	g_object_class_override_property(o_class,
			PROP_STATE, "state");
	g_object_class_override_property(o_class,
			PROP_IP4_ADDRESS, "ip4-address");
	g_object_class_override_property(o_class,
			PROP_ACTIVE_AP, "active-ap");

	g_object_class_install_property(o_class,
					PROP_SCAN_INTERVAL,
					g_param_spec_uint("scan-interval",
							  "Scan interval",
							  "Scan interval",
							  0, G_MAXUINT32,
							  0,
							  G_PARAM_READWRITE));

	g_type_class_add_private(klass, sizeof(DiversityNmPrivate));
}

static void diversity_nm_init(DiversityNm *nm)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	priv->iface = NULL;
	priv->state = DIVERSITY_NET_STATE_UNKNOWN;
	priv->ip4_address = 0;
}

static gint ap_equal(DiversityAp *a, DiversityAp *b)
{
	return 1;
}

static void on_scanned_ap(DiversityIwlib *iwlib, GSList *aps, DiversityNm *nm)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);
	GSList *tmp_list = aps;
	GSList *added = NULL, *old = NULL, *removed = priv->aps;

	while (tmp_list) {
		DiversityAp *ap = tmp_list->data;
		GSList *tmp;

		tmp = g_slist_find_custom(removed, ap,
			(GCompareFunc) ap_equal);

		if (tmp) {
			removed = g_slist_remove_link(removed, tmp);
			old = g_slist_concat(tmp, old);
		} else {
			g_object_ref(ap);
			added = g_slist_prepend(added, ap);
		}

		tmp_list = tmp_list->next;
	}

	if (removed) {
		g_signal_emit(nm, nm_signals[AP_REMOVED], 0,
				removed);

		g_slist_foreach(removed, (GFunc) g_object_unref, NULL);
		g_slist_free(removed);
	}

	if (added)
		g_signal_emit(nm, nm_signals[AP_ADDED], 0,
				added);

	priv->aps = g_slist_concat(old, added);
}

static DiversityNetState iwlib_state_map(guint iwlib_state)
{
	DiversityNetState s;

	switch (iwlib_state) {
	case DIVERSITY_IWLIB_STATE_DISCONNECTED:
		s = DIVERSITY_NET_STATE_DISCONNECTED;
		break;
	case DIVERSITY_IWLIB_STATE_ASSOCIATING:
		s = DIVERSITY_NET_STATE_CONFIG;
		break;
	case DIVERSITY_IWLIB_STATE_ASSOCIATED:
		s = DIVERSITY_NET_STATE_IP_CONFIG;
		break;
	default:
		s = DIVERSITY_NET_STATE_UNKNOWN;
		break;
	}

	return s;
}

static gboolean fail_dhcp(gpointer data)
{
	DiversityNm *nm = data;
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	g_debug("dhcp failed");

	diversity_nm_set_state(nm, DIVERSITY_NET_STATE_FAILED);

	priv->dhcp = 0;

	return FALSE;
}


#include <netinet/in.h>
#include <string.h>
#include <net/if.h>
#include <unistd.h>
#include <sys/ioctl.h>

guint
get_ip4_address(const gchar *iface)
{
	struct ifreq req;
	guint addr;
	gint fd, err;

	fd = socket(PF_INET, SOCK_DGRAM, 0);
	if (fd < 0)
		return 0;

	memset(&req, 0, sizeof(struct ifreq));
	strncpy(req.ifr_name, iface, IFNAMSIZ);

	err = ioctl(fd, SIOCGIFADDR, &req);

	close(fd);

	if (err < 0)
		return 0;

	addr = ((struct sockaddr_in *) &(req.ifr_addr))->sin_addr.s_addr;

	g_debug("%s has IP: %d.%d.%d.%d", iface,
			(addr) & 0xff,
			(addr >> 8) & 0xff,
			(addr >> 16) & 0xff,
			(addr >> 24) & 0xff);

	return addr;
}

static void do_dhcp(DiversityNm *nm)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);
	guint addr;

	addr = get_ip4_address(priv->iface);

	if (addr) {
		diversity_nm_set_ip4_address(nm, addr);
		diversity_nm_set_state(nm, DIVERSITY_NET_STATE_ACTIVATED);
	} else {
		priv->dhcp = g_timeout_add_seconds(5, fail_dhcp, nm);
	}
}

static void on_state_changed(DiversityIwlib *iwlib, guint state, DiversityNm *nm)
{
	state = iwlib_state_map(state);
	diversity_nm_set_state(nm, state);

	if (state == DIVERSITY_NET_STATE_IP_CONFIG)
		do_dhcp(nm);
}

static void diversity_nm_enum_devices(DiversityNm *nm)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	priv->iwlib = diversity_iwlib_new("eth0");
	if (!priv->iwlib)
		priv->iwlib = diversity_iwlib_new("wlan0");
}

DiversityNm *diversity_nm_new(void)
{
	DiversityNm *nm;
	DiversityNmPrivate *priv;
	const gchar *iface;
	guint state;

	nm = g_object_new(DIVERSITY_TYPE_NM,
			    "name", "nm",
			    NULL);

	priv = DIVERSITY_NM_GET_PRIVATE(nm);
	diversity_nm_enum_devices(nm);

	if (!priv->iwlib) {
		g_warning("failed to find any device");

		g_object_unref(nm);

		return NULL;
	}

	iface = diversity_iwlib_get_interface(priv->iwlib);
	priv->iface = g_strdup(iface);

	state = diversity_iwlib_get_state(priv->iwlib);
	priv->active_ap = diversity_iwlib_get_associate(priv->iwlib);
	if (priv->active_ap)
		g_object_ref(priv->active_ap);

	g_signal_connect(priv->iwlib, "state-changed",
			G_CALLBACK(on_state_changed), nm);
	g_signal_connect(priv->iwlib, "scanned-ap",
			G_CALLBACK(on_scanned_ap), nm);

	on_state_changed(priv->iwlib, state, nm);

	return nm;
}

const GSList *diversity_nm_list_aps(DiversityNm *nm)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	return priv->aps;
}

static gboolean diversity_nm_scan(gpointer data)
{
	DiversityNm *nm = data;
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	g_debug("request scan");

	diversity_iwlib_request_scan(priv->iwlib);

	return TRUE;
}

void diversity_nm_set_scan_interval(DiversityNm *nm, guint interval)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	if (priv->scan_interval != interval) {
		priv->scan_interval = interval;

		if (priv->scan_id) {
			g_source_remove(priv->scan_id);
			priv->scan_id = 0;
		}

		priv->scan_interval = interval;
		if (interval)
			priv->scan_id =
				g_timeout_add_seconds(interval,
						      diversity_nm_scan,
						      nm);
	}
}

guint diversity_nm_get_scan_interval(DiversityNm *nm)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	return priv->scan_interval;
}

void diversity_nm_set_interface(DiversityNm *nm, const gchar *iface)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	g_free(priv->iface);
	priv->iface = g_strdup(iface);

	g_object_notify(G_OBJECT(nm), "interface");
}

const gchar *diversity_nm_get_interface(DiversityNm *nm)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	return priv->iface;
}

void diversity_nm_set_state(DiversityNm *nm, guint state)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	if (priv->state != state) {
		priv->state = state;

		g_object_notify(G_OBJECT(nm), "state");
	}
}

guint diversity_nm_get_state(DiversityNm *nm)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	return priv->state;
}

void diversity_nm_set_ip4_address(DiversityNm *nm, guint address)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	if (priv->ip4_address != address) {
		priv->ip4_address = address;

		g_object_notify(G_OBJECT(nm), "ip4_address");
	}
}

guint diversity_nm_get_ip4_address(DiversityNm *nm)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(nm);

	return priv->ip4_address;
}

/* DiversityNet method implementations */

static gboolean diversity_nm_connect_ap(DiversityNet *net, DiversityAp *ap, GError **error)
{
	DiversityNmPrivate *priv = DIVERSITY_NM_GET_PRIVATE(net);
	const gchar *ssid;

	ssid = diversity_ap_get_ssid(ap);
	if (!ssid) {
		g_set_error(error, 0, 0, "AP does not have ESSID");

		return FALSE;
	}

	if (priv->active_ap)
		g_object_unref(priv->active_ap);
	priv->active_ap = ap;
	g_object_ref(ap);

	diversity_iwlib_request_associate(priv->iwlib, ap);

	return TRUE;
}

static gboolean diversity_nm_connect_wired(DiversityNet *net, GError **error)
{
	g_set_error(error, 0, 0, "NOT IMPLEMENTED");

	return FALSE;
}

static void diversity_nm_iface_init(gpointer g_iface, gpointer iface_data)
{
	DiversityNetIface *iface = (DiversityNetIface *) g_iface;

	iface->connect_ap = diversity_nm_connect_ap;
	iface->connect_wired = diversity_nm_connect_wired;
}
