/* diversity-qtopia-sms.c - DiversityQtopiaSms
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <dbus/dbus-glib.h>
#include <diversity-equipment.h>
#include <diversity-sms.h>
#include <string.h>
#include <errno.h>
#include "diversity-qtopia-sms.h"
#include "diversity-qtopia.h"
#include "qtopia-marshal.h"

enum {
	TAG_ADDED,
	LAST_SIGNAL
};

#define DIVERSITY_QTOPIA_SMS_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_QTOPIA_SMS, DiversityQtopiaSmsPrivate))
typedef struct _DiversityQtopiaSmsPrivate DiversityQtopiaSmsPrivate;

struct _DiversityQtopiaSmsPrivate {
	DBusGConnection *connection;
	DBusGProxy *proxy;

	guint open_interval;
	guint open_max_retry;

	guint open_id;
	guint open_retry;
	gboolean opened;
};

static guint sms_signals[LAST_SIGNAL] = { 0 };

static void diversity_qtopia_sms_iface_init(gpointer g_iface, gpointer iface_data);

G_DEFINE_TYPE_EXTENDED(DiversityQtopiaSms, diversity_qtopia_sms, DIVERSITY_TYPE_EQUIPMENT, 0,
		G_IMPLEMENT_INTERFACE(DIVERSITY_TYPE_SMS, diversity_qtopia_sms_iface_init));

static void diversity_qtopia_sms_dispose(GObject *obj)
{
	DiversityQtopiaSmsPrivate *priv =
		DIVERSITY_QTOPIA_SMS_GET_PRIVATE(obj);

	if (priv->open_id) {
		g_source_remove(priv->open_id);
		priv->open_id = 0;
	}

	if (priv->proxy) {
		if (priv->opened) {
			dbus_g_proxy_call(priv->proxy,
					  "close", NULL,
					  G_TYPE_INVALID,
					  G_TYPE_INVALID);

			priv->opened = FALSE;
		}

		g_object_unref(priv->proxy);
		priv->proxy = NULL;
	}

	((GObjectClass *) diversity_qtopia_sms_parent_class)->dispose(obj);
}

static void diversity_qtopia_sms_finalize(GObject *obj)
{
	DiversityQtopiaSmsPrivate *priv =
		DIVERSITY_QTOPIA_SMS_GET_PRIVATE(obj);

	if (priv->connection)
		dbus_g_connection_unref(priv->connection);

	((GObjectClass *) diversity_qtopia_sms_parent_class)->finalize(obj);
}

static gboolean diversity_qtopia_sms_equip(DiversityEquipment *eqp)
{
	return TRUE;
}

static void diversity_qtopia_sms_unequip(DiversityEquipment *eqp)
{
}

static gboolean diversity_qtopia_sms_set_config(DiversityEquipment *eqp, const gchar *key, const GValue *val)
{
	return FALSE;
}

static gboolean diversity_qtopia_sms_get_config(DiversityEquipment *eqp, const gchar *key, GValue *val)
{
	return FALSE;
}

static void diversity_qtopia_sms_class_init(DiversityQtopiaSmsClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;
	DiversityEquipmentClass *eqp_class = (DiversityEquipmentClass *) klass;

	o_class->dispose = diversity_qtopia_sms_dispose;
	o_class->finalize = diversity_qtopia_sms_finalize;

	eqp_class->equip = diversity_qtopia_sms_equip;
	eqp_class->unequip = diversity_qtopia_sms_unequip;

	eqp_class->set_config = diversity_qtopia_sms_set_config;
	eqp_class->get_config = diversity_qtopia_sms_get_config;

	g_type_class_add_private(klass, sizeof(DiversityQtopiaSmsPrivate));

	sms_signals[TAG_ADDED] =
		g_signal_new("tag-added",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityQtopiaSmsClass, tag_added),
			     NULL, NULL,
			     qtopia_marshal_VOID__STRING_DOUBLE_DOUBLE_STRING,
			     G_TYPE_NONE, 4,
			     G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE,
			     G_TYPE_DOUBLE,
			     G_TYPE_DOUBLE,
			     G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);
}

static void diversity_qtopia_sms_init(DiversityQtopiaSms *qsms)
{
	DiversityQtopiaSmsPrivate *priv =
		DIVERSITY_QTOPIA_SMS_GET_PRIVATE(qsms);

	priv->connection = NULL;
	priv->proxy = NULL;

	priv->open_interval = 5;
	priv->open_max_retry = 12;

	priv->open_id = 0;
	priv->open_retry = 0;
	priv->opened = FALSE;
}

DiversityQtopiaSms *diversity_qtopia_sms_new(const gchar *name)
{
	if (!name)
		return NULL;

	return g_object_new(DIVERSITY_TYPE_QTOPIA_SMS,
			    "name", name,
			    NULL);
}

#define L6N_MIN_LEN (3 + 6)
static gboolean parse_l6n(GHashTable *dict, gdouble *lon, gdouble *lat, gchar **desc)
{
	GValue *val;
	gchar *body;
	gchar *p, *endp;

	val = g_hash_table_lookup(dict, "content");
	body = (gchar *) g_value_get_string(val);

	if (!body || strlen(body) < L6N_MIN_LEN)
		return FALSE;

	body = g_strdup(body);

	p = body;

	if (strncmp(p, "l6n(", 4) != 0)
		goto fail;
	p += 4;

	errno = 0;

	*lat = g_ascii_strtod(p, &endp);
	if (!endp || errno || *lat < -90.0 || *lat >= 90.0)
		goto fail;
	p = endp;

	if (*p != ',')
		goto fail;
	p++;

	*lon = g_ascii_strtod(p, &endp);
	if (!endp || errno || *lon < -180.0 || *lon >= 180.0)
		goto fail;
	p = endp;

	if (strncmp(p, ")\n", 2) != 0)
		goto fail;
	p += 2;

	*desc = g_strdup(p);

	g_free(body);

	return TRUE;
fail:
	g_free(body);

	return FALSE;
}

static gboolean diversity_qtopia_sms_parse(DiversityQtopiaSms *qsms, const gchar *id)
{
	DiversityQtopiaSmsPrivate *priv =
		DIVERSITY_QTOPIA_SMS_GET_PRIVATE(qsms);
	GError *error = NULL;
	GType out_type;
	GHashTable *dict;
	gdouble lon, lat;
	gchar *desc;

	out_type = dbus_g_type_get_map(
			"GHashTable", G_TYPE_STRING, G_TYPE_VALUE);

	if (!dbus_g_proxy_call(priv->proxy,
			       "message", &error,
			       G_TYPE_STRING, id,
			       G_TYPE_INVALID,
			       out_type, &dict,
			       G_TYPE_INVALID)) {
		g_warning("failed to query message %s: %s",
				id, error->message);

		g_error_free(error);

		return FALSE;
	}

	if (!parse_l6n(dict, &lon, &lat, &desc)) {
		g_hash_table_destroy(dict);

		return FALSE;
	}

	g_signal_emit(qsms, sms_signals[TAG_ADDED], 0,
			id, lon, lat, desc);

	g_free(desc);
	g_hash_table_destroy(dict);

	return TRUE;
}

static void on_new_sms(DBusGProxy *proxy, GPtrArray *ids, gpointer user_data)
{
	DiversityQtopiaSms *qsms = user_data;
	guint i;

	for (i = 0; i < ids->len; i++) {
		GValue *val = g_ptr_array_index(ids, i);
		const gchar *id;

		id = g_value_get_string(val);

		g_debug("receive message %s", id);

		if (!diversity_qtopia_sms_parse(qsms, id))
			g_debug("failed to parse message %s", id);
	}
}

static gboolean timeout_open(gpointer data)
{
	DiversityQtopiaSms *qsms = data;
	DiversityQtopiaSmsPrivate *priv =
		DIVERSITY_QTOPIA_SMS_GET_PRIVATE(qsms);

	if (priv->opened) {
		priv->open_id = 0;

		return FALSE;
	}

	if (dbus_g_proxy_call(priv->proxy,
			      "open", NULL,
			      G_TYPE_INVALID,
			      G_TYPE_INVALID)) {
		GType out_type;
		GPtrArray *ids;

		g_debug("sms opened");

		out_type = dbus_g_type_get_collection("GPtrArray",
						      G_TYPE_VALUE);

		dbus_g_proxy_add_signal(priv->proxy,
					"newSms",
					out_type,
					G_TYPE_INVALID);
		dbus_g_proxy_connect_signal(priv->proxy,
					    "newSms",
					    G_CALLBACK(on_new_sms),
					    qsms, NULL);

		if (dbus_g_proxy_call(priv->proxy,
				      "listMessages", NULL,
				      G_TYPE_INVALID,
				      out_type, &ids,
				      G_TYPE_INVALID)) {
			guint i;

			g_debug("list messages");

			on_new_sms(priv->proxy, ids, qsms);

			for (i = 0; i < ids->len; i++) {
				GValue *val = g_ptr_array_index(ids, i);

				g_value_unset(val);
			}

			g_ptr_array_free(ids, TRUE);
		}

		priv->opened = TRUE;
		priv->open_id = 0;

		return FALSE;
	} else {
		priv->open_retry++;

		if (priv->open_retry >= priv->open_max_retry) {
			g_warning("failed to open sms");

			priv->open_id = 0;

			return FALSE;
		}
	}

	return TRUE;
}

gboolean diversity_qtopia_sms_open_and_start(DiversityQtopiaSms *qsms)
{
	DiversityQtopiaSmsPrivate *priv =
		DIVERSITY_QTOPIA_SMS_GET_PRIVATE(qsms);
	GError *error = NULL;

	if (priv->opened || priv->open_id)
		return TRUE;

	priv->connection = dbus_g_bus_get(QTOPIA_DBUS_BUS, &error);
	if (!priv->connection) {
		g_warning("failed to get bus: %s", error->message);

		g_error_free(error);

		return FALSE;
	}

	priv->proxy =
		dbus_g_proxy_new_for_name(priv->connection,
					  QTOPIA_DBUS_SERVICE,
					  QTOPIA_SMS_DBUS_PATH,
					  QTOPIA_SMS_DBUS_INTERFACE);

	if (!priv->proxy) {
		g_warning("failed to get proxy");

		dbus_g_connection_unref(priv->connection);
		priv->connection = NULL;

		return FALSE;
	}

	if (timeout_open(qsms))
		priv->open_id = g_timeout_add_seconds(priv->open_interval,
				timeout_open, qsms);

	return TRUE;
}

gboolean diversity_qtopia_sms_remove(DiversityQtopiaSms *qsms, const gchar *uid)
{
	DiversityQtopiaSmsPrivate *priv =
		DIVERSITY_QTOPIA_SMS_GET_PRIVATE(qsms);
	GError *error = NULL;
	gboolean ret;

	if (!priv->opened || !uid)
		return FALSE;

	g_debug("remove message %s", uid);

	ret = dbus_g_proxy_call(priv->proxy, "removeMessage", &error,
				G_TYPE_STRING, uid,
				G_TYPE_INVALID,
				G_TYPE_BOOLEAN, NULL,
				G_TYPE_INVALID);
	if (!ret) {
		g_debug("failed to remove message %s: %s",
				uid, error->message);

		g_error_free(error);
	}

	return ret;
}

/* DiversitySms method implementations */

static DiversitySmsStatus diversity_qtopia_sms_get_status(DiversitySms *sms)
{
	DiversityQtopiaSmsPrivate *priv =
		DIVERSITY_QTOPIA_SMS_GET_PRIVATE(sms);

	return (priv->opened)
		? DIVERSITY_SMS_STATUS_READY
		: DIVERSITY_SMS_STATUS_NOTREADY;
}

static gboolean diversity_qtopia_sms_send(DiversitySms *sms, const gchar *number, const gchar *message, gboolean ask_ds, GError **error)
{
	DiversityQtopiaSmsPrivate *priv =
		DIVERSITY_QTOPIA_SMS_GET_PRIVATE(sms);
	gboolean ret;

	if (!priv->opened) {
		g_set_error(error, 0, 0, "SMS is not opened");

		return FALSE;
	}

	ret = dbus_g_proxy_call(priv->proxy, "send", error,
				G_TYPE_STRING, number,
				G_TYPE_STRING, message,
				G_TYPE_BOOLEAN, ask_ds,
				G_TYPE_INVALID,
				G_TYPE_STRING, NULL,
				G_TYPE_INVALID);

	return ret;
}

static void diversity_qtopia_sms_iface_init(gpointer g_iface, gpointer iface_data)
{
	DiversitySmsIface *iface = (DiversitySmsIface *) g_iface;

	iface->get_status = diversity_qtopia_sms_get_status;
	iface->send = diversity_qtopia_sms_send;
}
