/* diversity-conf.c -
 *
 * Copyright 2008 OpenMoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <string.h>
#include <diversity-conf.h>

struct _DiversityConf {
	GKeyFile *key_file;
	gchar *group;
};

DiversityConf *diversity_conf_new(void)
{
	DiversityConf *conf;

	conf = g_new(DiversityConf, 1);
	conf->key_file = g_key_file_new();
	conf->group = NULL;

	return conf;
}

void diversity_conf_free(DiversityConf *conf)
{
	g_key_file_free(conf->key_file);
	g_free(conf->group);

	g_free(conf);
}

gboolean diversity_conf_save(DiversityConf *conf, const gchar *path)
{
	gchar *data;
	gboolean ret;

	data = g_key_file_to_data(conf->key_file, NULL, NULL);
	if (!data)
		return FALSE;

	//g_print("%s\n", data);

	ret = g_file_set_contents(path, data, -1, NULL);
	g_free(data);

	return ret;
}

gboolean diversity_conf_load(DiversityConf *conf, const gchar *path)
{
	return g_key_file_load_from_file(conf->key_file,
			path, G_KEY_FILE_NONE, NULL);
}

void diversity_conf_use_group(DiversityConf *conf, const gchar *group)
{
	g_free(conf->group);
	conf->group = g_strdup(group);
}

gchar **diversity_conf_list_groups(DiversityConf *conf, gsize *len)
{
	return g_key_file_get_groups(conf->key_file, len);
}

void diversity_conf_set_integer(DiversityConf *conf, const gchar *key, gint v)
{
	g_key_file_set_integer(conf->key_file, conf->group, key, v);
}

gboolean diversity_conf_get_integer(DiversityConf *conf, const gchar *key, gint *v)
{
	GError *error = NULL;
	gint tmp;

	tmp = g_key_file_get_integer(conf->key_file, conf->group,
			key, &error);

	if (error) {
		g_error_free(error);

		return FALSE;
	}

	if (v)
		*v = tmp;

	return TRUE;
}

void diversity_conf_set_boolean(DiversityConf *conf, const gchar *key, gboolean v)
{
	g_key_file_set_boolean(conf->key_file, conf->group, key, v);
}

gboolean diversity_conf_get_boolean(DiversityConf *conf, const gchar *key, gboolean *v)
{
	GError *error = NULL;
	gboolean tmp;

	tmp = g_key_file_get_boolean(conf->key_file, conf->group,
			key, &error);

	if (error) {
		g_error_free(error);

		return FALSE;
	}

	if (v)
		*v = tmp;

	return TRUE;
}

void diversity_conf_set_string(DiversityConf *conf, const gchar *key, const gchar *v)
{
	g_key_file_set_string(conf->key_file, conf->group, key, v);
}

gboolean diversity_conf_get_string(DiversityConf *conf, const gchar *key, gchar **v)
{
	GError *error = NULL;
	gchar *tmp;

	tmp = g_key_file_get_string(conf->key_file, conf->group,
			key, &error);

	if (error) {
		g_error_free(error);

		return FALSE;
	}

	if (v)
		*v = tmp;

	return TRUE;
}

void diversity_conf_set_rectangle(DiversityConf *conf, const gchar *key, const DiversityRectangle *rect)
{
	gdouble v[4];

	v[0] = rect->x;
	v[1] = rect->y;
	v[2] = rect->width;
	v[3] = rect->height;

	g_key_file_set_double_list(conf->key_file, conf->group,
			key, v, 4);
}

gboolean diversity_conf_get_rectangle(DiversityConf *conf, const gchar *key, DiversityRectangle *rect)
{
	gdouble *v;
	gsize len;

	v = g_key_file_get_double_list(conf->key_file, conf->group, key, &len, NULL);
	if (!v || len != 4) {
		g_free(v);

		return FALSE;
	}

	if (rect) {
		rect->x = v[0];
		rect->y = v[1];
		rect->width = v[2];
		rect->height = v[3];
	}

	g_free(v);

	return TRUE;
}

typedef struct {
	DiversityConf *conf;
	const gchar *prefix;
} DiversityConfDictData;

static void set_dict(gpointer key, gpointer val, gpointer user_data)
{
	DiversityConfDictData *dict_data = user_data;
	gchar *full;

	full = g_strconcat(dict_data->prefix, "_", key, NULL);
	g_key_file_set_string(dict_data->conf->key_file,
			      dict_data->conf->group,
			      full, val);
	g_free(full);
}

void diversity_conf_set_dictionary(DiversityConf *conf, const gchar *key, GHashTable *dict)
{
	DiversityConfDictData dict_data;

	dict_data.conf = conf;
	dict_data.prefix = key;

	g_hash_table_foreach(dict, set_dict, &dict_data);
}

gboolean diversity_conf_get_dictionary(DiversityConf *conf, const gchar *key, GHashTable *dict)
{
	gchar **keys;
	gsize num_keys, i, len;

	len = strlen(key);

	keys = g_key_file_get_keys(conf->key_file, conf->group,
			&num_keys, NULL);

	for (i = 0; i < num_keys; i++) {
		gchar *val;

		if (strncmp(keys[i], key, len) != 0 ||
		    keys[i][len] != '_' ||
		    keys[i][len + 1] == '\0')
			continue;

		val = g_key_file_get_string(conf->key_file, conf->group,
				keys[i], NULL);

		if (!dict) {
			g_free(val);

			break;
		}

		if (val)
			g_hash_table_insert(dict,
					g_strdup(keys[i] + len + 1), val);
	}

	g_strfreev(keys);

	return TRUE;
}
